"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Literal, TypedDict, Union

from typing_extensions import NotRequired

from .channel import ChannelTypeWithoutThread, ThreadMetadata
from .member import Member
from .message import Attachment
from .role import Role
from .snowflake import Snowflake
from .threads import ThreadType
from .user import User

if TYPE_CHECKING:
    from .message import Message


InteractionType = Literal[1, 2, 3, 4, 5]


class _BasePartialChannel(TypedDict):
    id: Snowflake
    name: str
    permissions: str


class PartialChannel(_BasePartialChannel):
    type: ChannelTypeWithoutThread


class PartialThread(_BasePartialChannel):
    type: ThreadType
    thread_metadata: ThreadMetadata
    parent_id: Snowflake


class ResolvedData(TypedDict, total=False):
    users: Dict[str, User]
    members: Dict[str, Member]
    roles: Dict[str, Role]
    channels: Dict[str, Union[PartialChannel, PartialThread]]
    messages: Dict[str, Message]
    attachments: Dict[str, Attachment]


class _BaseApplicationCommandInteractionDataOption(TypedDict):
    name: str


class _CommandGroupApplicationCommandInteractionDataOption(
    _BaseApplicationCommandInteractionDataOption
):
    type: Literal[1, 2]
    options: List[ApplicationCommandInteractionDataOption]


class _BaseValueApplicationCommandInteractionDataOption(
    _BaseApplicationCommandInteractionDataOption, total=False
):
    focused: bool


class _StringValueApplicationCommandInteractionDataOption(
    _BaseValueApplicationCommandInteractionDataOption
):
    type: Literal[3]
    value: str


class _IntegerValueApplicationCommandInteractionDataOption(
    _BaseValueApplicationCommandInteractionDataOption
):
    type: Literal[4]
    value: int


class _BooleanValueApplicationCommandInteractionDataOption(
    _BaseValueApplicationCommandInteractionDataOption
):
    type: Literal[5]
    value: bool


class _SnowflakeValueApplicationCommandInteractionDataOption(
    _BaseValueApplicationCommandInteractionDataOption
):
    type: Literal[6, 7, 8, 9, 11]
    value: Snowflake


class _NumberValueApplicationCommandInteractionDataOption(
    _BaseValueApplicationCommandInteractionDataOption
):
    type: Literal[10]
    value: float


_ValueApplicationCommandInteractionDataOption = Union[
    _StringValueApplicationCommandInteractionDataOption,
    _IntegerValueApplicationCommandInteractionDataOption,
    _BooleanValueApplicationCommandInteractionDataOption,
    _SnowflakeValueApplicationCommandInteractionDataOption,
    _NumberValueApplicationCommandInteractionDataOption,
]


ApplicationCommandInteractionDataOption = Union[
    _CommandGroupApplicationCommandInteractionDataOption,
    _ValueApplicationCommandInteractionDataOption,
]


class _BaseApplicationCommandInteractionData(TypedDict):
    id: Snowflake
    name: str
    resolved: NotRequired[ResolvedData]
    guild_id: NotRequired[Snowflake]


class ChatInputApplicationCommandInteractionData(
    _BaseApplicationCommandInteractionData, total=False
):
    type: Literal[1]
    options: List[ApplicationCommandInteractionDataOption]


class _BaseNonChatInputApplicationCommandInteractionData(
    _BaseApplicationCommandInteractionData
):
    target_id: Snowflake


class UserApplicationCommandInteractionData(
    _BaseNonChatInputApplicationCommandInteractionData
):
    type: Literal[2]


class MessageApplicationCommandInteractionData(
    _BaseNonChatInputApplicationCommandInteractionData
):
    type: Literal[3]


ApplicationCommandInteractionData = Union[
    ChatInputApplicationCommandInteractionData,
    UserApplicationCommandInteractionData,
    MessageApplicationCommandInteractionData,
]


class _BaseMessageComponentInteractionData(TypedDict):
    custom_id: str


class ButtonMessageComponentInteractionData(_BaseMessageComponentInteractionData):
    component_type: Literal[2]


class SelectMessageComponentInteractionData(_BaseMessageComponentInteractionData):
    component_type: Literal[3]
    values: List[str]


MessageComponentInteractionData = Union[
    ButtonMessageComponentInteractionData, SelectMessageComponentInteractionData
]


class ModalSubmitTextInputInteractionData(TypedDict):
    type: Literal[4]
    custom_id: str
    value: str


ModalSubmitComponentItemInteractionData = ModalSubmitTextInputInteractionData


class ModalSubmitActionRowInteractionData(TypedDict):
    type: Literal[1]
    components: List[ModalSubmitComponentItemInteractionData]


ModalSubmitComponentInteractionData = Union[
    ModalSubmitActionRowInteractionData, ModalSubmitComponentItemInteractionData
]


class ModalSubmitInteractionData(TypedDict):
    custom_id: str
    components: List[ModalSubmitComponentInteractionData]


InteractionData = Union[
    ApplicationCommandInteractionData,
    MessageComponentInteractionData,
    ModalSubmitInteractionData,
]


class _BaseInteraction(TypedDict):
    id: Snowflake
    application_id: Snowflake
    token: str
    version: Literal[1]
    guild_id: NotRequired[Snowflake]
    channel_id: NotRequired[Snowflake]
    locale: NotRequired[str]
    guild_locale: NotRequired[str]


class PingInteraction(_BaseInteraction):
    type: Literal[1]


class ApplicationCommandInteraction(_BaseInteraction):
    type: Literal[2, 4]
    data: ApplicationCommandInteractionData


class MessageComponentInteraction(_BaseInteraction):
    type: Literal[3]
    data: MessageComponentInteractionData


class ModalSubmitInteraction(_BaseInteraction):
    type: Literal[5]
    data: ModalSubmitInteractionData


Interaction = Union[
    PingInteraction,
    ApplicationCommandInteraction,
    MessageComponentInteraction,
    ModalSubmitInteraction,
]


class MessageInteraction(TypedDict):
    id: Snowflake
    type: InteractionType
    name: str
    user: User
    member: NotRequired[Member]



from discord_self._vendor.discord.types.application import PartialApplication
from discord_self._vendor.discord.types.channel import PartialChannel
from discord_self._vendor.discord.types.guild import InviteGuild
from discord_self._vendor.discord.types.invite import (
    GatewayInvite,
    GatewayInviteCreate,
    GatewayInviteDelete,
    IncompleteInvite,
    Invite,
    InviteTargetType,
    InviteWithCounts,
    VanityInvite,
)
from discord_self._vendor.discord.types.scheduled_event import GuildScheduledEvent
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.user import PartialUser

__all__ = [
    "GatewayInvite",
    "GatewayInviteCreate",
    "GatewayInviteDelete",
    "GuildScheduledEvent",
    "IncompleteInvite",
    "Invite",
    "InviteGuild",
    "InviteTargetType",
    "InviteWithCounts",
    "PartialApplication",
    "PartialChannel",
    "PartialUser",
    "Snowflake",
    "VanityInvite",
]



"""
Discord API Wrapper
~~~~~~~~~~~~~~~~~~~

A basic wrapper for the Discord user API.

:copyright: (c) 2015-present Rapptz and 2021-present Dolfies
:license: MIT, see LICENSE for more details.
"""

__title__ = "discord.py-self"
__author__ = "Dolfies"
__license__ = "MIT"
__copyright__ = "Copyright 2015-present Rapptz and 2021-present Dolfies"
__version__ = "2.0.1"

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import logging
from typing import Literal, NamedTuple

from . import abc as abc
from . import opus as opus
from . import utils as utils
from .activity import *
from .affinity import *
from .application import *
from .asset import *
from .audit_logs import *
from .automod import *
from .billing import *
from .calls import *
from .channel import *
from .client import *
from .colour import *
from .commands import *
from .components import *
from .connections import *
from .embeds import *
from .emoji import *
from .entitlements import *
from .enums import *
from .errors import *
from .file import *
from .flags import *
from .guild import *
from .guild_premium import *
from .handlers import *
from .integrations import *
from .interactions import *
from .invite import *
from .library import *
from .member import *
from .mentions import *
from .message import *
from .metadata import *
from .modal import *
from .object import *
from .partial_emoji import *
from .payments import *
from .permissions import *
from .player import *
from .profile import *
from .promotions import *
from .raw_models import *
from .reaction import *
from .relationship import *
from .role import *
from .scheduled_event import *
from .settings import *
from .stage_instance import *
from .sticker import *
from .store import *
from .subscriptions import *
from .team import *
from .template import *
from .threads import *
from .tracking import *
from .user import *
from .voice_client import *
from .webhook import *
from .welcome_screen import *
from .widget import *


class _VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int


version_info: _VersionInfo = _VersionInfo(
    major=2, minor=0, micro=1, releaselevel="final", serial=0
)

logging.getLogger(__name__).addHandler(logging.NullHandler())


del logging, NamedTuple, Literal, _VersionInfo



from discord_self._vendor.discord import discord
from discord_self._vendor.discord.__main__ import (
    add_newbot_args,
    add_newcog_args,
    core,
    main,
    newbot,
    newcog,
    parse_args,
    show_version,
    to_path,
)

__all__ = [
    "add_newbot_args",
    "add_newcog_args",
    "core",
    "discord",
    "main",
    "newbot",
    "newcog",
    "parse_args",
    "show_version",
    "to_path",
]



from discord_self._vendor.discord.types.guild import Guild
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.template import CreateTemplate, Template
from discord_self._vendor.discord.types.user import User

__all__ = ["CreateTemplate", "Guild", "Snowflake", "Template", "User"]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple, Union

from .enums import RelationshipAction, RelationshipType, Status, try_enum
from .mixins import Hashable
from .object import Object
from .utils import MISSING, parse_time

if TYPE_CHECKING:
    from datetime import datetime

    from typing_extensions import Self

    from .activity import ActivityTypes
    from .state import ConnectionState, Presence
    from .types.gateway import RelationshipEvent
    from .types.user import Relationship as RelationshipPayload
    from .user import User

# fmt: off
__all__ = (
    'Relationship',
)
# fmt: on


class Relationship(Hashable):
    """Represents a relationship in Discord.

    A relationship is like a friendship, a person who is blocked, etc.

    .. container:: operations

        .. describe:: x == y

            Checks if two relationships are equal.

        .. describe:: x != y

            Checks if two relationships are not equal.

        .. describe:: hash(x)

            Return the relationship's hash.

    Attributes
    -----------
    user: :class:`User`
        The user you have the relationship with.
    type: :class:`RelationshipType`
        The type of relationship you have.
    nick: Optional[:class:`str`]
        The user's friend nickname (if applicable).

        .. versionadded:: 1.9

        .. versionchanged:: 2.0
            Renamed ``nickname`` to :attr:`nick`.
    since: Optional[:class:`datetime.datetime`]
        When the relationship was created.
        Only available for type :class:`RelationshipType.incoming_request`.

        .. versionadded:: 2.0
    """

    __slots__ = ("_presence", "since", "nick", "type", "user", "_state")

    if TYPE_CHECKING:
        user: User

    def __init__(self, *, state: ConnectionState, data: RelationshipPayload) -> None:
        self._state = state
        self._presence: Optional[Presence] = None
        self._update(data)

    def _update(self, data: Union[RelationshipPayload, RelationshipEvent]) -> None:
        self.type: RelationshipType = try_enum(RelationshipType, data["type"])
        self.nick: Optional[str] = data.get("nickname")
        self.since: Optional[datetime] = parse_time(data.get("since"))

        if not getattr(self, "user", None):
            if "user" in data:
                self.user = self._state.store_user(data["user"])  # type: ignore
            else:
                user_id = int(data["id"])
                self.user = self._state.get_user(user_id) or Object(id=user_id)  # type: ignore # Lying for better developer UX

    @classmethod
    def _from_implicit(cls, *, state: ConnectionState, user: User) -> Relationship:
        self = cls.__new__(cls)
        self._state = state
        self._presence = None
        self.type = RelationshipType.implicit
        self.nick = None
        self.since = None
        self.user = user
        return self

    @classmethod
    def _copy(cls, relationship: Self, presence: Presence) -> Self:
        self = cls.__new__(cls)  # to bypass __init__

        self._state = relationship._state
        self._presence = presence
        self.type = relationship.type
        self.nick = relationship.nick
        self.since = relationship.since
        self.user = relationship.user
        return self

    def __repr__(self) -> str:
        return (
            f"<Relationship user={self.user!r} type={self.type!r} nick={self.nick!r}>"
        )

    @property
    def id(self) -> int:
        """:class:`int`: Returns the relationship's ID."""
        return self.user.id

    @property
    def presence(self) -> Presence:
        state = self._state
        return (
            self._presence
            or state._presences.get(self.user.id)
            or state.create_offline_presence()
        )

    @property
    def status(self) -> Status:
        """:class:`Status`: The user's overall status.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.
        """
        return try_enum(Status, self.presence.client_status.status)

    @property
    def raw_status(self) -> str:
        """:class:`str`: The user's overall status as a string value.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.
        """
        return self.presence.client_status.status

    @property
    def mobile_status(self) -> Status:
        """:class:`Status`: The user's status on a mobile device, if applicable.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.
        """
        return try_enum(Status, self.presence.client_status.mobile or "offline")

    @property
    def desktop_status(self) -> Status:
        """:class:`Status`: The user's status on the desktop client, if applicable.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.
        """
        return try_enum(Status, self.presence.client_status.desktop or "offline")

    @property
    def web_status(self) -> Status:
        """:class:`Status`: The user's status on the web client, if applicable.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.
        """
        return try_enum(Status, self.presence.client_status.web or "offline")

    def is_on_mobile(self) -> bool:
        """:class:`bool`: A helper function that determines if a user is active on a mobile device.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.
        """
        return self.presence.client_status.mobile is not None

    @property
    def activities(self) -> Tuple[ActivityTypes, ...]:
        """Tuple[Union[:class:`BaseActivity`, :class:`Spotify`]]: Returns the activities that
        the user is currently doing.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.

        .. note::

            Due to a Discord API limitation, a user's Spotify activity may not appear
            if they are listening to a song with a title longer
            than 128 characters. See :issue:`1738` for more information.
        """
        return self.presence.activities

    @property
    def activity(self) -> Optional[ActivityTypes]:
        """Optional[Union[:class:`BaseActivity`, :class:`Spotify`]]: Returns the primary
        activity the user is currently doing. Could be ``None`` if no activity is being done.

        .. versionadded:: 2.0

        .. note::

            This is only reliably provided for type :class:`RelationshipType.friend`.

        .. note::

            Due to a Discord API limitation, this may be ``None`` if
            the user is listening to a song on Spotify with a title longer
            than 128 characters. See :issue:`1738` for more information.

        .. note::

            A user may have multiple activities, these can be accessed under :attr:`activities`.
        """
        if self.activities:
            return self.activities[0]

    async def delete(self) -> None:
        """|coro|

        Deletes the relationship.

        Depending on the type, this could mean unfriending or unblocking the user,
        denying an incoming friend request, discarding an outgoing friend request, etc.

        Raises
        ------
        HTTPException
            Deleting the relationship failed.
        """
        action = RelationshipAction.deny_request
        if self.type is RelationshipType.friend:
            action = RelationshipAction.unfriend
        elif self.type is RelationshipType.blocked:
            action = RelationshipAction.unblock
        elif self.type is RelationshipType.incoming_request:
            action = RelationshipAction.deny_request
        elif self.type is RelationshipType.outgoing_request:
            action = RelationshipAction.remove_pending_request

        await self._state.http.remove_relationship(self.user.id, action=action)

    async def accept(self) -> None:
        """|coro|

        Accepts the relationship request. Only applicable for
        type :class:`RelationshipType.incoming_request`.

        Raises
        -------
        HTTPException
            Accepting the relationship failed.
        """
        await self._state.http.add_relationship(
            self.user.id, action=RelationshipAction.accept_request
        )

    async def edit(self, nick: Optional[str] = MISSING) -> None:
        """|coro|

        Edits the relationship.

        .. versionadded:: 1.9

        .. versionchanged:: 2.0
            Changed the name of the method to :meth:`edit`.
            The edit is no longer in-place.

        Parameters
        ----------
        nick: Optional[:class:`str`]
            The nickname to change to. Can be ``None`` to denote no nickname.

        Raises
        -------
        HTTPException
            Changing the nickname failed.
        """
        payload = {}
        if nick is not MISSING:
            payload["nickname"] = nick

        await self._state.http.edit_relationship(self.user.id, **payload)



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional

from .enums import EntitlementType, GiftStyle, PremiumType, try_enum
from .flags import GiftFlags
from .mixins import Hashable
from .payments import EntitlementPayment
from .promotions import Promotion
from .store import SKU, StoreListing, SubscriptionPlan
from .subscriptions import Subscription, SubscriptionTrial
from .utils import _get_as_snowflake, parse_time, utcnow

if TYPE_CHECKING:
    from datetime import datetime

    from .abc import Snowflake
    from .guild import Guild
    from .state import ConnectionState
    from .types.entitlements import (
        Entitlement as EntitlementPayload,
    )
    from .types.entitlements import (
        Gift as GiftPayload,
    )
    from .types.entitlements import (
        GiftBatch as GiftBatchPayload,
    )
    from .user import User

__all__ = (
    "Entitlement",
    "Gift",
    "GiftBatch",
)


class Entitlement(Hashable):
    """Represents a Discord entitlement.

    .. container:: operations

        .. describe:: x == y

            Checks if two entitlements are equal.

        .. describe:: x != y

            Checks if two entitlements are not equal.

        .. describe:: hash(x)

            Returns the entitlement's hash.

        .. describe:: bool(x)

            Checks if the entitlement is active.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the entitlement.
    type: :class:`EntitlementType`
        The type of entitlement.
    user_id: :class:`int`
        The ID of the user the entitlement is for.
    sku_id: :class:`int`
        The ID of the SKU the entitlement grants.
    application_id: :class:`int`
        The ID of the application that owns the SKU the entitlement grants.
    promotion_id: Optional[:class:`int`]
        The ID of the promotion the entitlement is from.
    parent_id: Optional[:class:`int`]
        The ID of the entitlement's parent.
    guild_id: Optional[:class:`int`]
        The ID of the guild the entitlement is for.
    branches: List[:class:`int`]
        The IDs of the branches the entitlement grants.
    gifter_id: Optional[:class:`int`]
        The ID of the user that gifted the entitlement.
    gift_style: Optional[:class:`GiftStyle`]
        The style of the gift attached to this entitlement.
    gift_batch_id: Optional[:class:`int`]
        The ID of the batch the gift attached to this entitlement is from.
    deleted: :class:`bool`
        Whether the entitlement is deleted.
    consumed: :class:`bool`
        Whether the entitlement is consumed.
    starts_at: Optional[:class:`datetime.datetime`]
        When the entitlement period starts.
    ends_at: Optional[:class:`datetime.datetime`]
        When the entitlement period ends.
    subscription_id: Optional[:class:`int`]
        The ID of the subscription the entitlement is from.
    subscription_plan: Optional[:class:`SubscriptionPlan`]
        The subscription plan the entitlement is for.

        .. note::

            This is a partial object without price information.
    sku: Optional[:class:`SKU`]
        The SKU the entitlement grants.
    payment: Optional[:class:`EntitlementPayment`]
        The payment made for the entitlement.
        Not available in some contexts.
    """

    __slots__ = (
        "id",
        "type",
        "user_id",
        "sku_id",
        "application_id",
        "promotion_id",
        "parent_id",
        "guild_id",
        "branches",
        "gifter_id",
        "gift_style",
        "gift_batch_id",
        "_gift_flags",
        "deleted",
        "consumed",
        "starts_at",
        "ends_at",
        "subscription_id",
        "subscription_plan",
        "sku",
        "payment",
        "_state",
    )

    def __init__(self, *, data: EntitlementPayload, state: ConnectionState):
        self._state = state
        self._update(data)

    def _update(self, data: EntitlementPayload):
        state = self._state

        self.id: int = int(data["id"])
        self.type: EntitlementType = try_enum(EntitlementType, data["type"])
        self.user_id: int = int(data.get("user_id") or state.self_id)  # type: ignore
        self.sku_id: int = int(data["sku_id"])
        self.application_id: int = int(data["application_id"])
        self.promotion_id: Optional[int] = _get_as_snowflake(data, "promotion_id")
        self.parent_id: Optional[int] = _get_as_snowflake(data, "parent_id")
        self.guild_id: Optional[int] = _get_as_snowflake(data, "guild_id")
        self.branches: List[int] = [int(branch) for branch in data.get("branches", [])]
        self.gifter_id: Optional[int] = _get_as_snowflake(data, "gifter_user_id")
        self.gift_style: Optional[GiftStyle] = try_enum(
            GiftStyle, data.get("gift_style")
        )
        self.gift_batch_id: Optional[int] = _get_as_snowflake(
            data, "gift_code_batch_id"
        )
        self._gift_flags: int = data.get("gift_code_flags", 0)

        self.deleted: bool = data.get("deleted", False)
        self.consumed: bool = data.get("consumed", False)
        self.starts_at: Optional[datetime] = parse_time(data.get("starts_at"))
        self.ends_at: Optional[datetime] = parse_time(data.get("ends_at"))

        self.subscription_id: Optional[int] = _get_as_snowflake(data, "subscription_id")
        self.subscription_plan: Optional[SubscriptionPlan] = (
            SubscriptionPlan(data=data["subscription_plan"], state=state)
            if "subscription_plan" in data
            else None
        )
        self.sku: Optional[SKU] = (
            SKU(data=data["sku"], state=state) if "sku" in data else None
        )
        self.payment: Optional[EntitlementPayment] = (
            EntitlementPayment(data=data["payment"], entitlement=self)
            if "payment" in data
            else None
        )

    def __repr__(self) -> str:
        return f"<Entitlement id={self.id} type={self.type!r} sku_id={self.sku_id} application_id={self.application_id}>"

    def __bool__(self) -> bool:
        return self.is_active()

    @property
    def guild(self) -> Optional[Guild]:
        """:class:`Guild`: Returns the guild the entitlement is for, if accessible."""
        return self._state._get_guild(self.guild_id)

    @property
    def premium_type(self) -> Optional[PremiumType]:
        """Optional[:class:`PremiumType`]: The premium type this entitlement grants, if it is for a premium subscription."""
        return PremiumType.from_sku_id(self.sku_id)

    @property
    def gift_flags(self) -> GiftFlags:
        """:class:`GiftFlags`: Returns the flags for the gift this entitlement is attached to."""
        return GiftFlags._from_value(self._gift_flags)

    def is_giftable(self) -> bool:
        """:class:`bool`: Whether the entitlement is giftable."""
        return self.type == EntitlementType.user_gift and not self.gifter_id

    def is_active(self) -> bool:
        """:class:`bool`: Whether the entitlement is active and offering perks."""
        # This is a copy of the logic used in the client

        if self.is_giftable() or self.deleted:
            return False  # Giftable entitlements have not yet been gifted therefore are not active
        if self.starts_at and self.starts_at > utcnow():
            return False  # Entitlement has not started yet
        if self.ends_at and self.ends_at < utcnow():
            return False  # Entitlement has ended

        if self.type == EntitlementType.premium_subscription:
            # Premium subscription entitlements are only active
            # if the SKU is offered for free to premium subscribers
            # and the user is a premium subscriber
            sku = self.sku
            if sku and not sku.premium:
                return False
            if (
                self._state.user
                and not self._state.user.premium_type == PremiumType.nitro
            ):
                return False

        return True

    async def subscription(self) -> Optional[Subscription]:
        """|coro|

        Retrieves the subscription this entitlement is attached to, if applicable.

        Raises
        ------
        NotFound
            You cannot access this subscription.
        HTTPException
            Fetching the subscription failed.

        Returns
        -------
        Optional[:class:`Subscription`]
            The retrieved subscription, if applicable.
        """
        if not self.subscription_id:
            return

        data = await self._state.http.get_subscription(self.subscription_id)
        return Subscription(data=data, state=self._state)

    async def consume(self) -> None:
        """|coro|

        Consumes the entitlement. This marks a given user entitlement as expended,
        and removes the entitlement from the user's active entitlements.

        This should be called after the user has received the relevant item,
        and only works on entitlements for SKUs of type :attr:`SKUType.consumable`.

        Raises
        ------
        Forbidden
            You do not have permissions to access this application.
        HTTPException
            Consuming the entitlement failed.
        """
        await self._state.http.consume_app_entitlement(self.application_id, self.id)

    async def delete(self) -> None:
        """|coro|

        Deletes the entitlement. This removes the entitlement from the user's
        entitlements, and is irreversible.

        This is only useable on entitlements of type :attr:`EntitlementType.test_mode_purchase`.

        Raises
        ------
        Forbidden
            You do not have permissions to access this application.
        HTTPException
            Deleting the entitlement failed.
        """
        await self._state.http.delete_app_entitlement(self.application_id, self.id)


class Gift:
    """Represents a Discord gift.

    .. container:: operations

        .. describe:: x == y

            Checks if two gifts are equal.

        .. describe:: x != y

            Checks if two gifts are not equal.

        .. describe:: hash(x)

            Returns the gift's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    code: :class:`str`
        The gift's code.
    expires_at: Optional[:class:`datetime.datetime`]
        When the gift expires.
    application_id: Optional[:class:`int`]
        The ID of the application that owns the SKU the gift is for.
        Not available in all contexts.
    batch_id: Optional[:class:`int`]
        The ID of the batch the gift is from.
    sku_id: :class:`int`
        The ID of the SKU the gift is for.
    entitlement_branches: List[:class:`int`]
        A list of entitlements the gift is for.
    gift_style: Optional[:class:`GiftStyle`]
        The style of the gift.
    max_uses: :class:`int`
        The maximum number of times the gift can be used.
    uses: :class:`int`
        The number of times the gift has been used.
    redeemed: :class:`bool`
        Whether the user has redeemed the gift.
    revoked: :class:`bool`
        Whether the gift has been revoked.
    guild_id: Optional[:class:`int`]
        The ID of the guild the gift was redeemed in.
        Not available in all contexts.
    channel_id: Optional[:class:`int`]
        The ID of the channel the gift was redeemed in.
        Not available in all contexts.
    store_listing: Optional[:class:`StoreListing`]
        The store listing for the SKU the gift is for.
        Not available in all contexts.
    promotion: Optional[:class:`Promotion`]
        The promotion the gift is a part of, if any.
    subscription_trial: Optional[:class:`SubscriptionTrial`]
        The subscription trial the gift is a part of, if any.
    subscription_plan_id: Optional[:class:`int`]
        The ID of the subscription plan the gift is for, if any.
    subscription_plan: Optional[:class:`SubscriptionPlan`]
        The subscription plan the gift is for, if any.
    user: Optional[:class:`User`]
        The user who created the gift, if applicable.
    """

    __slots__ = (
        "code",
        "expires_at",
        "application_id",
        "batch_id",
        "sku_id",
        "entitlement_branches",
        "gift_style",
        "_flags",
        "max_uses",
        "uses",
        "redeemed",
        "revoked",
        "guild_id",
        "channel_id",
        "store_listing",
        "promotion",
        "subscription_trial",
        "subscription_plan_id",
        "subscription_plan",
        "user",
        "_state",
    )

    def __init__(self, *, data: GiftPayload, state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def _update(self, data: GiftPayload) -> None:
        state = self._state

        self.code: str = data["code"]
        self.expires_at: Optional[datetime] = parse_time(data.get("expires_at"))
        self.application_id: Optional[int] = _get_as_snowflake(data, "application_id")
        self.batch_id: Optional[int] = _get_as_snowflake(data, "batch_id")
        self.subscription_plan_id: Optional[int] = _get_as_snowflake(
            data, "subscription_plan_id"
        )
        self.sku_id: int = int(data["sku_id"])
        self.entitlement_branches: List[int] = [
            int(x) for x in data.get("entitlement_branches", [])
        ]
        self.gift_style: Optional[GiftStyle] = (
            try_enum(GiftStyle, data["gift_style"]) if data.get("gift_style") else None
        )  # type: ignore
        self._flags: int = data.get("flags", 0)

        self.max_uses: int = data.get("max_uses", 0)
        self.uses: int = data.get("uses", 0)
        self.redeemed: bool = data.get("redeemed", False)
        self.revoked: bool = data.get("revoked", False)

        self.guild_id: Optional[int] = _get_as_snowflake(data, "guild_id")
        self.channel_id: Optional[int] = _get_as_snowflake(data, "channel_id")

        self.store_listing: Optional[StoreListing] = (
            StoreListing(data=data["store_listing"], state=state)
            if "store_listing" in data
            else None
        )
        self.promotion: Optional[Promotion] = (
            Promotion(data=data["promotion"], state=state)
            if "promotion" in data
            else None
        )
        self.subscription_trial: Optional[SubscriptionTrial] = (
            SubscriptionTrial(data["subscription_trial"])
            if "subscription_trial" in data
            else None
        )
        self.subscription_plan: Optional[SubscriptionPlan] = (
            SubscriptionPlan(data=data["subscription_plan"], state=state)
            if "subscription_plan" in data
            else None
        )
        self.user: Optional[User] = (
            self._state.create_user(data["user"]) if "user" in data else None
        )

    def __repr__(self) -> str:
        return f"<Gift code={self.code!r} sku_id={self.sku_id} uses={self.uses} max_uses={self.max_uses} redeemed={self.redeemed}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Gift) and other.code == self.code

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, Gift):
            return other.code != self.code
        return True

    def __hash__(self) -> int:
        return hash(self.code)

    @property
    def id(self) -> str:
        """:class:`str`: Returns the code portion of the gift."""
        return self.code

    @property
    def url(self) -> str:
        """:class:`str`: Returns the gift's URL."""
        return f"https://discord.gift/{self.code}"

    @property
    def remaining_uses(self) -> int:
        """:class:`int`: Returns the number of remaining uses for the gift."""
        return self.max_uses - self.uses

    @property
    def flags(self) -> GiftFlags:
        """:class:`GiftFlags`: Returns the gift's flags."""
        return GiftFlags._from_value(self._flags)

    @property
    def premium_type(self) -> Optional[PremiumType]:
        """Optional[:class:`PremiumType`]: The premium type this gift grants, if it is for a premium subscription."""
        return PremiumType.from_sku_id(self.sku_id) if self.is_subscription() else None

    def is_claimed(self) -> bool:
        """:class:`bool`: Checks if the gift has been used up."""
        return self.uses >= self.max_uses if self.max_uses else False

    def is_expired(self) -> bool:
        """:class:`bool`: Checks if the gift has expired."""
        return self.expires_at < utcnow() if self.expires_at else False

    def is_subscription(self) -> bool:
        """:class:`bool`: Checks if the gift is for a subscription."""
        return self.subscription_plan_id is not None

    def is_premium_subscription(self) -> bool:
        """:class:`bool`: Checks if the gift is for a premium subscription."""
        return (
            self.is_subscription()
            and self.application_id == self._state.premium_subscriptions_application.id
        )

    async def redeem(
        self,
        payment_source: Optional[Snowflake] = None,
        *,
        channel: Optional[Snowflake] = None,
        gateway_checkout_context: Optional[str] = None,
    ) -> Entitlement:
        """|coro|

        Redeems the gift.

        Parameters
        ----------
        payment_source: Optional[:class:`PaymentSource`]
            The payment source to use for the redemption.
            Only required if the gift's :attr:`flags` have :attr:`GiftFlags.payment_source_required` set to ``True``.
        channel: Optional[Union[:class:`TextChannel`, :class:`VoiceChannel`, :class:`StageChannel`, :class:`Thread`, :class:`DMChannel`, :class:`GroupChannel`]]
            The channel to redeem the gift in. This is usually the channel the gift was sent in.
            While this is optional, it is recommended to pass this in.
        gateway_checkout_context: Optional[:class:`str`]
            The current checkout context.

        Raises
        ------
        HTTPException
            The gift failed to redeem.

        Returns
        -------
        :class:`Entitlement`
            The entitlement that was created from redeeming the gift.
        """
        data = await self._state.http.redeem_gift(
            self.code,
            payment_source.id if payment_source else None,
            channel.id if channel else None,
            gateway_checkout_context,
        )
        return Entitlement(data=data, state=self._state)

    async def delete(self) -> None:
        """|coro|

        Revokes the gift.

        This is only possible for gifts the current account has created.

        Raises
        ------
        NotFound
            The owned gift was not found.
        HTTPException
            The gift failed to delete.
        """
        await self._state.http.delete_gift(self.code)


class GiftBatch(Hashable):
    """Represents a batch of gifts for an SKU.

    .. container:: operations

        .. describe:: x == y

            Checks if two gift batches are equal.

        .. describe:: x != y

            Checks if two gift batches are not equal.

        .. describe:: hash(x)

            Returns the gift batch's hash.

        .. describe:: str(x)

            Returns the gift batch's description.

    .. versionadded:: 2.0

    Attributes
    -----------
    id: :class:`int`
        The ID of the gift batch.
    application_id: :class:`int`
        The ID of the application the gift batch is for.
    sku_id: :class:`int`
        The ID of the SKU the gift batch is for.
    amount: :class:`int`
        The amount of gifts in the batch.
    description: :class:`str`
        The description of the gift batch.
    entitlement_branches: List[:class:`int`]
        The entitlement branches the gift batch is for.
    entitlement_starts_at: Optional[:class:`datetime.datetime`]
        When the entitlement is valid from.
    entitlement_ends_at: Optional[:class:`datetime.datetime`]
        When the entitlement is valid until.
    """

    __slots__ = (
        "id",
        "application_id",
        "sku_id",
        "amount",
        "description",
        "entitlement_branches",
        "entitlement_starts_at",
        "entitlement_ends_at",
        "_state",
    )

    def __init__(
        self, *, data: GiftBatchPayload, state: ConnectionState, application_id: int
    ) -> None:
        self._state: ConnectionState = state
        self.id: int = int(data["id"])
        self.application_id = application_id
        self.sku_id: int = int(data["sku_id"])
        self.amount: int = data["amount"]
        self.description: str = data.get("description", "")
        self.entitlement_branches: List[int] = [
            int(branch) for branch in data.get("entitlement_branches", [])
        ]
        self.entitlement_starts_at: Optional[datetime] = parse_time(
            data.get("entitlement_starts_at")
        )
        self.entitlement_ends_at: Optional[datetime] = parse_time(
            data.get("entitlement_ends_at")
        )

    def __repr__(self) -> str:
        return f"<GiftBatch id={self.id} sku_id={self.sku_id} amount={self.amount} description={self.description!r}>"

    def __str__(self) -> str:
        return self.description

    def is_valid(self) -> bool:
        """:class:`bool`: Checks if the gift batch is valid."""
        if self.entitlement_starts_at and self.entitlement_starts_at > utcnow():
            return False
        if self.entitlement_ends_at and self.entitlement_ends_at < utcnow():
            return False
        return True

    async def download(self) -> bytes:
        """|coro|

        Returns the gifts in the gift batch in CSV format.

        Raises
        -------
        Forbidden
            You do not have permissions to download the batch.
        HTTPException
            Downloading the batch failed.

        Returns
        -------
        :class:`bytes`
            The report content.
        """
        return await self._state.http.get_gift_batch_csv(self.application_id, self.id)



from discord_self._vendor.discord.abc import (
    Connectable,
    GuildChannel,
    Messageable,
    PrivateChannel,
    Snowflake,
    User,
)
from discord_self._vendor.discord.commands import ApplicationCommand

__all__ = [
    "ApplicationCommand",
    "Connectable",
    "GuildChannel",
    "Messageable",
    "PrivateChannel",
    "Snowflake",
    "User",
]



from discord_self._vendor.discord.types.channel import PrivacyLevel
from discord_self._vendor.discord.types.member import Member
from discord_self._vendor.discord.types.scheduled_event import (
    EntityMetadata,
    EntityType,
    EventStatus,
    ExternalScheduledEvent,
    GuildScheduledEvent,
    GuildScheduledEventWithUserCount,
    ScheduledEventUser,
    ScheduledEventUsers,
    ScheduledEventUsersWithMember,
    ScheduledEventUserWithMember,
    StageInstanceScheduledEvent,
    VoiceScheduledEvent,
)
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.user import User

__all__ = [
    "EntityMetadata",
    "EntityType",
    "EventStatus",
    "ExternalScheduledEvent",
    "GuildScheduledEvent",
    "GuildScheduledEventWithUserCount",
    "Member",
    "PrivacyLevel",
    "ScheduledEventUser",
    "ScheduledEventUserWithMember",
    "ScheduledEventUsers",
    "ScheduledEventUsersWithMember",
    "Snowflake",
    "StageInstanceScheduledEvent",
    "User",
    "VoiceScheduledEvent",
]



from discord_self._vendor.discord.types.application import (
    EULA,
    Achievement,
    ActiveDeveloperResponse,
    ActiveDeveloperWebhook,
    ActivityStatistics,
    Application,
    ApplicationDiscoverability,
    ApplicationInstallParams,
    Asset,
    BaseAchievement,
    BaseApplication,
    Branch,
    BranchSize,
    Build,
    BuildFile,
    ClientPlatformConfig,
    Company,
    CreatedBuild,
    CreatedBuildFile,
    DownloadSignature,
    EmbeddedActivityConfig,
    EmbeddedActivityPlatform,
    GlobalActivityStatistics,
    IntegrationApplication,
    Manifest,
    ManifestLabel,
    MetadataApplication,
    PartialApplication,
    PartialRoleConnection,
    RoleConnection,
    RoleConnectionMetadata,
    StoreAsset,
    Ticket,
    WhitelistedUser,
)
from discord_self._vendor.discord.types.guild import PartialGuild
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.team import Team
from discord_self._vendor.discord.types.user import PartialUser

__all__ = [
    "Achievement",
    "ActiveDeveloperResponse",
    "ActiveDeveloperWebhook",
    "ActivityStatistics",
    "Application",
    "ApplicationDiscoverability",
    "ApplicationInstallParams",
    "Asset",
    "BaseAchievement",
    "BaseApplication",
    "Branch",
    "BranchSize",
    "Build",
    "BuildFile",
    "ClientPlatformConfig",
    "Company",
    "CreatedBuild",
    "CreatedBuildFile",
    "DownloadSignature",
    "EULA",
    "EmbeddedActivityConfig",
    "EmbeddedActivityPlatform",
    "GlobalActivityStatistics",
    "IntegrationApplication",
    "Manifest",
    "ManifestLabel",
    "MetadataApplication",
    "PartialApplication",
    "PartialGuild",
    "PartialRoleConnection",
    "PartialUser",
    "RoleConnection",
    "RoleConnectionMetadata",
    "Snowflake",
    "StoreAsset",
    "Team",
    "Ticket",
    "WhitelistedUser",
]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from functools import reduce
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    overload,
)

from .enums import UserFlags

if TYPE_CHECKING:
    from typing_extensions import Self


__all__ = (
    "Capabilities",
    "SystemChannelFlags",
    "MessageFlags",
    "PublicUserFlags",
    "PrivateUserFlags",
    "MemberCacheFlags",
    "ApplicationFlags",
    "ChannelFlags",
    "PremiumUsageFlags",
    "PurchasedFlags",
    "PaymentSourceFlags",
    "SKUFlags",
    "PaymentFlags",
    "PromotionFlags",
    "GiftFlags",
    "LibraryApplicationFlags",
    "ApplicationDiscoveryFlags",
    "FriendSourceFlags",
    "FriendDiscoveryFlags",
    "HubProgressFlags",
    "OnboardingProgressFlags",
    "AutoModPresets",
    "MemberFlags",
)

BF = TypeVar("BF", bound="BaseFlags")


class flag_value:
    def __init__(self, func: Callable[[Any], int]):
        self.flag: int = func(None)
        self.__doc__: Optional[str] = func.__doc__

    @overload
    def __get__(self, instance: None, owner: Type[BF]) -> Self: ...

    @overload
    def __get__(self, instance: BF, owner: Type[BF]) -> bool: ...

    def __get__(self, instance: Optional[BF], owner: Type[BF]) -> Any:
        if instance is None:
            return self
        return instance._has_flag(self.flag)

    def __set__(self, instance: BaseFlags, value: bool) -> None:
        instance._set_flag(self.flag, value)

    def __repr__(self) -> str:
        return f"<flag_value flag={self.flag!r}>"


class alias_flag_value(flag_value):
    pass


def fill_with_flags(*, inverted: bool = False) -> Callable[[Type[BF]], Type[BF]]:
    def decorator(cls: Type[BF]) -> Type[BF]:
        # fmt: off
        cls.VALID_FLAGS = {
            name: value.flag
            for name, value in cls.__dict__.items()
            if isinstance(value, flag_value)
        }
        # fmt: on

        if inverted:
            max_bits = max(cls.VALID_FLAGS.values()).bit_length()
            cls.DEFAULT_VALUE = -1 + (2**max_bits)
        else:
            cls.DEFAULT_VALUE = 0

        return cls

    return decorator


# Flags must inherit from this and use the decorator above
class BaseFlags:
    VALID_FLAGS: ClassVar[Dict[str, int]]
    DEFAULT_VALUE: ClassVar[int]

    value: int

    __slots__ = ("value",)

    def __init__(self, **kwargs: bool):
        self.value = self.DEFAULT_VALUE
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def _from_value(cls, value):
        self = cls.__new__(cls)
        self.value = value
        return self

    def __or__(self, other: Self) -> Self:
        return self._from_value(self.value | other.value)

    def __and__(self, other: Self) -> Self:
        return self._from_value(self.value & other.value)

    def __xor__(self, other: Self) -> Self:
        return self._from_value(self.value ^ other.value)

    def __ior__(self, other: Self) -> Self:
        self.value |= other.value
        return self

    def __iand__(self, other: Self) -> Self:
        self.value &= other.value
        return self

    def __ixor__(self, other: Self) -> Self:
        self.value ^= other.value
        return self

    def __invert__(self) -> Self:
        max_bits = max(self.VALID_FLAGS.values()).bit_length()
        max_value = -1 + (2**max_bits)
        return self._from_value(self.value ^ max_value)

    def __bool__(self) -> bool:
        return self.value != self.DEFAULT_VALUE

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.value == other.value

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} value={self.value}>"

    def __iter__(self) -> Iterator[Tuple[str, bool]]:
        for name, value in self.__class__.__dict__.items():
            if isinstance(value, alias_flag_value):
                continue

            if isinstance(value, flag_value):
                yield (name, self._has_flag(value.flag))

    def _has_flag(self, o: int) -> bool:
        return (self.value & o) == o

    def _set_flag(self, o: int, toggle: bool) -> None:
        if toggle is True:
            self.value |= o
        elif toggle is False:
            self.value &= ~o
        else:
            raise TypeError(
                f"Value to set for {self.__class__.__name__} must be a bool."
            )


class ArrayFlags(BaseFlags):
    @classmethod
    def _from_value(cls: Type[Self], value: List[int]) -> Self:
        self = cls.__new__(cls)
        self.value = reduce(lambda a, b: a | (1 << b - 1), value, 0)
        return self

    def to_array(self) -> List[int]:
        return [i + 1 for i in range(self.value.bit_length()) if self.value & (1 << i)]


@fill_with_flags()
class Capabilities(BaseFlags):
    """Wraps up the Discord gateway capabilities.

    Capabilities are used to determine what gateway features a client support.

    This is meant to be used internally by the library.

    .. container:: operations

        .. describe:: x == y

            Checks if two capabilities are equal.
        .. describe:: x != y

            Checks if two capabilities are not equal.
        .. describe:: x | y, x |= y

            Returns a capabilities instance with all enabled flags from
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x & y, x &= y

            Returns a capabilities instance with only flags enabled on
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x ^ y, x ^= y

            Returns a capabilities instance with only flags enabled on
            only one of x or y, not on both.

            .. versionadded:: 2.0
        .. describe:: ~x

            Returns a capabilities instance with all flags inverted from x.

            .. versionadded:: 2.0
        .. describe:: hash(x)

               Return the capability's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    # The unfortunate thing about capabilities is that while a lot of these options
    # may be useful to the library (i.e. to expose to users for customization),
    # we match the official client's values for anti-spam purposes :(

    @classmethod
    def default(cls: Type[Self]) -> Self:
        """Returns a :class:`Capabilities` with the current value used by the library."""
        return cls._from_value(8189)

    @flag_value
    def lazy_user_notes(self):
        """:class:`bool`: Disable preloading of user notes in READY."""
        return 1 << 0

    @flag_value
    def no_affine_user_ids(self):
        """:class:`bool`: Disable implicit relationship updates."""
        return 1 << 1

    @flag_value
    def versioned_read_states(self):
        """:class:`bool`: Enable versioned read states (change READY ``read_state`` to an object with ``version``/``partial``)."""
        return 1 << 2

    @flag_value
    def versioned_user_guild_settings(self):
        """:class:`bool`: Enable versioned user guild settings (change READY ``user_guild_settings`` to an object with ``version``/``partial``)."""
        return 1 << 3

    @flag_value
    def dedupe_user_objects(self):
        """:class:`bool`: Enable dehydration of the READY payload (move all user objects to a ``users`` array and replace them in various places in the READY payload with ``user_id`` or ``recipient_id``, move member object(s) from initial guild objects to ``merged_members``)."""
        return 1 << 4

    @flag_value
    def prioritized_ready_payload(self):
        """:class:`bool`: Enable prioritized READY payload (enable READY_SUPPLEMENTAL, move ``voice_states`` and ``embedded_activities`` from initial guild objects and ``merged_presences`` from READY, as well as split ``merged_members`` and (sometimes) ``private_channels``/``lazy_private_channels`` between the events)."""
        # Requires self.dedupe_user_objects
        return 1 << 5 | 1 << 4

    @flag_value
    def multiple_guild_experiment_populations(self):
        """:class:`bool`: Handle multiple guild experiment populations (change the fourth entry of arrays in the ``guild_experiments`` array in READY to have an array of population arrays)."""
        return 1 << 6

    @flag_value
    def non_channel_read_states(self):
        """:class:`bool`: Handle non-channel read states (change READY ``read_state`` to include read states tied to server events, server home, and the mobile notification center)."""
        return 1 << 7

    @flag_value
    def auth_token_refresh(self):
        """:class:`bool`: Enable auth token refresh (add ``auth_token?`` to READY; this is sent when Discord wants to change the client's token, and was used for the mfa. token migration)."""
        return 1 << 8

    @flag_value
    def user_settings_proto(self):
        """:class:`bool`: Disable legacy user settings (remove ``user_settings`` from READY and stop sending USER_SETTINGS_UPDATE)."""
        return 1 << 9

    @flag_value
    def client_state_v2(self):
        """:class:`bool`: Enable client caching v2 (move guild properties in guild objects to a ``properties`` subkey and add ``data_mode`` and ``version`` to the objects, as well as change ``client_state`` in IDENTIFY)."""
        return 1 << 10

    @flag_value
    def passive_guild_update(self):
        """:class:`bool`: Enable passive guild update (replace ``CHANNEL_UNREADS_UPDATE`` with ``PASSIVE_UPDATE_V1``, a similar event that includes a ``voice_states`` array and a ``members`` array that includes the members of aforementioned voice states)."""
        return 1 << 11

    @flag_value
    def unknown_12(self):
        """:class:`bool`: Unknown."""
        return 1 << 12


@fill_with_flags(inverted=True)
class SystemChannelFlags(BaseFlags):
    r"""Wraps up a Discord system channel flag value.

    Similar to :class:`Permissions`\, the properties provided are two way.
    You can set and retrieve individual bits using the properties as if they
    were regular bools. This allows you to edit the system flags easily.

    To construct an object you can pass keyword arguments denoting the flags
    to enable or disable.

    .. container:: operations

        .. describe:: x == y

            Checks if two SystemChannelFlags are equal.
        .. describe:: x != y

            Checks if two SystemChannelFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a SystemChannelFlags instance with all enabled flags from
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x & y, x &= y

            Returns a SystemChannelFlags instance with only flags enabled on
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x ^ y, x ^= y

            Returns a SystemChannelFlags instance with only flags enabled on
            only one of x or y, not on both.

            .. versionadded:: 2.0
        .. describe:: ~x

            Returns a SystemChannelFlags instance with all flags inverted from x.

            .. versionadded:: 2.0
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    # For some reason the flags for system channels are "inverted"
    # Ergo, if they're set then it means "suppress" (off in the GUI toggle)
    # Since this is counter-intuitive from an API perspective and annoying
    # these will be inverted automatically

    def _has_flag(self, o: int) -> bool:
        return (self.value & o) != o

    def _set_flag(self, o: int, toggle: bool) -> None:
        if toggle is True:
            self.value &= ~o
        elif toggle is False:
            self.value |= o
        else:
            raise TypeError("Value to set for SystemChannelFlags must be a bool")

    @flag_value
    def join_notifications(self):
        """:class:`bool`: Returns ``True`` if the system channel is used for member join notifications."""
        return 1

    @flag_value
    def premium_subscriptions(self):
        """:class:`bool`: Returns ``True`` if the system channel is used for "Nitro boosting" notifications."""
        return 2

    @flag_value
    def guild_reminder_notifications(self):
        """:class:`bool`: Returns ``True`` if the system channel is used for server setup helpful tips notifications.

        .. versionadded:: 2.0
        """
        return 4

    @flag_value
    def join_notification_replies(self):
        """:class:`bool`: Returns ``True`` if sticker reply button ("Wave to say hi!") is
        shown for member join notifications.

        .. versionadded:: 2.0
        """
        return 8

    @flag_value
    def role_subscription_purchase_notifications(self):
        """:class:`bool`: Returns ``True`` if role subscription purchase and renewal
        notifications are enabled.

        .. versionadded:: 2.0
        """
        return 16

    @flag_value
    def role_subscription_purchase_notification_replies(self):
        """:class:`bool`: Returns ``True`` if the role subscription notifications
        have a sticker reply button.

        .. versionadded:: 2.0
        """
        return 32


@fill_with_flags()
class MessageFlags(BaseFlags):
    r"""Wraps up a Discord Message flag value.

    See :class:`SystemChannelFlags`.

    .. container:: operations

        .. describe:: x == y

            Checks if two MessageFlags are equal.
        .. describe:: x != y

            Checks if two MessageFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a MessageFlags instance with all enabled flags from
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x & y, x &= y

            Returns a MessageFlags instance with only flags enabled on
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x ^ y, x ^= y

            Returns a MessageFlags instance with only flags enabled on
            only one of x or y, not on both.

            .. versionadded:: 2.0
        .. describe:: ~x

            Returns a MessageFlags instance with all flags inverted from x.

            .. versionadded:: 2.0
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.

    .. versionadded:: 1.3

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def crossposted(self):
        """:class:`bool`: Returns ``True`` if the message is the original crossposted message."""
        return 1

    @flag_value
    def is_crossposted(self):
        """:class:`bool`: Returns ``True`` if the message was crossposted from another channel."""
        return 2

    @flag_value
    def suppress_embeds(self):
        """:class:`bool`: Returns ``True`` if the message's embeds have been suppressed."""
        return 4

    @flag_value
    def source_message_deleted(self):
        """:class:`bool`: Returns ``True`` if the source message for this crosspost has been deleted."""
        return 8

    @flag_value
    def urgent(self):
        """:class:`bool`: Returns ``True`` if the message is an urgent message.

        An urgent message is one sent by Discord Trust and Safety.
        """
        return 16

    @flag_value
    def has_thread(self):
        """:class:`bool`: Returns ``True`` if the message is associated with a thread.

        .. versionadded:: 2.0
        """
        return 32

    @flag_value
    def ephemeral(self):
        """:class:`bool`: Returns ``True`` if the message is ephemeral.

        .. versionadded:: 2.0
        """
        return 64

    @flag_value
    def loading(self):
        """:class:`bool`: Returns ``True`` if the message is an interaction response and the bot
        is "thinking".

        .. versionadded:: 2.0
        """
        return 128

    @flag_value
    def failed_to_mention_some_roles_in_thread(self):
        """:class:`bool`: Returns ``True`` if the message failed to mention some roles in a thread
        and add their members to the thread.

        .. versionadded:: 2.0
        """
        return 256

    @flag_value
    def link_not_discord_warning(self):
        """:class:`bool`: Returns ``True`` if this message contains a link that impersonates
        Discord and should show a warning.

        .. versionadded:: 2.0
        """
        return 1024

    @flag_value
    def suppress_notifications(self):
        """:class:`bool`: Returns ``True`` if the message will not trigger push and desktop notifications.

        .. versionadded:: 2.0
        """
        return 4096

    @alias_flag_value
    def silent(self):
        """:class:`bool`: Alias for :attr:`suppress_notifications`.

        .. versionadded:: 2.0
        """
        return 4096

    @flag_value
    def is_voice_message(self):
        """:class:`bool`: Returns ``True`` if the message's audio attachments are rendered as voice messages.

        .. versionadded:: 2.0
        """
        return 8192


@fill_with_flags()
class PublicUserFlags(BaseFlags):
    r"""Wraps up the Discord User Public flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PublicUserFlags are equal.
        .. describe:: x != y

            Checks if two PublicUserFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PublicUserFlags instance with all enabled flags from
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x & y, x &= y

            Returns a PublicUserFlags instance with only flags enabled on
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x ^ y, x ^= y

            Returns a PublicUserFlags instance with only flags enabled on
            only one of x or y, not on both.

            .. versionadded:: 2.0
        .. describe:: ~x

            Returns a PublicUserFlags instance with all flags inverted from x.

            .. versionadded:: 2.0
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 1.4

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def staff(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Employee."""
        return UserFlags.staff.value

    @flag_value
    def partner(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Partner."""
        return UserFlags.partner.value

    @flag_value
    def hypesquad(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Events member."""
        return UserFlags.hypesquad.value

    @flag_value
    def bug_hunter(self):
        """:class:`bool`: Returns ``True`` if the user is a level 1 Bug Hunter."""
        return UserFlags.bug_hunter.value

    @alias_flag_value
    def bug_hunter_level_1(self):
        """:class:`bool`: An alias for :attr:`bug_hunter`.

        .. versionadded:: 2.0
        """
        return UserFlags.bug_hunter_level_1.value

    @flag_value
    def hypesquad_bravery(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Bravery member."""
        return UserFlags.hypesquad_bravery.value

    @flag_value
    def hypesquad_brilliance(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Brilliance member."""
        return UserFlags.hypesquad_brilliance.value

    @flag_value
    def hypesquad_balance(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Balance member."""
        return UserFlags.hypesquad_balance.value

    @flag_value
    def early_supporter(self):
        """:class:`bool`: Returns ``True`` if the user is an Early Supporter."""
        return UserFlags.early_supporter.value

    @flag_value
    def team_user(self):
        """:class:`bool`: Returns ``True`` if the user is a Team User."""
        return UserFlags.team_user.value

    @flag_value
    def system(self):
        """:class:`bool`: Returns ``True`` if the user is a system user (i.e. represents Discord officially)."""
        return UserFlags.system.value

    @flag_value
    def bug_hunter_level_2(self):
        """:class:`bool`: Returns ``True`` if the user is a level 2 Bug Hunter"""
        return UserFlags.bug_hunter_level_2.value

    @flag_value
    def verified_bot(self):
        """:class:`bool`: Returns ``True`` if the user is a Verified Bot."""
        return UserFlags.verified_bot.value

    @flag_value
    def verified_bot_developer(self):
        """:class:`bool`: Returns ``True`` if the user is an Early Verified Bot Developer."""
        return UserFlags.verified_bot_developer.value

    @alias_flag_value
    def early_verified_bot_developer(self):
        """:class:`bool`: An alias for :attr:`verified_bot_developer`.

        .. versionadded:: 1.5
        """
        return UserFlags.verified_bot_developer.value

    @flag_value
    def discord_certified_moderator(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Certified Moderator.

        .. versionadded:: 2.0
        """
        return UserFlags.discord_certified_moderator.value

    @flag_value
    def bot_http_interactions(self):
        """:class:`bool`: Returns ``True`` if the user is a bot that only uses HTTP interactions
        and is shown in the online member list.

        .. versionadded:: 2.0
        """
        return UserFlags.bot_http_interactions.value

    @flag_value
    def spammer(self):
        """:class:`bool`: Returns ``True`` if the user is flagged as a spammer by Discord.

        .. versionadded:: 2.0
        """
        return UserFlags.spammer.value

    @flag_value
    def active_developer(self):
        """:class:`bool`: Returns ``True`` if the user is an active developer.

        .. versionadded:: 2.0
        """
        return UserFlags.active_developer.value

    def all(self) -> List[UserFlags]:
        """List[:class:`UserFlags`]: Returns all flags the user has."""
        return [
            public_flag
            for public_flag in UserFlags
            if self._has_flag(public_flag.value)
        ]


@fill_with_flags()
class PrivateUserFlags(PublicUserFlags):
    r"""Wraps up the Discord User flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PrivateUserFlags are equal.
        .. describe:: x != y

            Checks if two PrivateUserFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PrivateUserFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a PrivateUserFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a PrivateUserFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a PrivateUserFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases or inherited flags are not shown.

    .. versionadded:: 2.0

    .. note::
        These are only available on your own user flags.

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def premium_promo_dismissed(self):
        """:class:`bool`: Returns ``True`` if the user has dismissed the current premium promotion."""
        return UserFlags.premium_promo_dismissed.value

    @flag_value
    def has_unread_urgent_messages(self):
        """:class:`bool`: Returns ``True`` if the user has unread urgent system messages."""
        return UserFlags.has_unread_urgent_messages.value

    @flag_value
    def mfa_sms(self):
        """:class:`bool`: Returns ``True`` if the user has SMS recovery for MFA enabled."""
        return UserFlags.mfa_sms.value

    @flag_value
    def underage_deleted(self):
        """:class:`bool`: Returns ``True`` if the user has been flagged for deletion for being underage."""
        return UserFlags.underage_deleted.value

    @flag_value
    def partner_or_verification_application(self):
        """:class:`bool`: Returns ``True`` if the user has a partner or a verification application."""
        return UserFlags.partner_or_verification_application.value

    @flag_value
    def disable_premium(self):
        """:class:`bool`: Returns ``True`` if the user bought premium but has it manually disabled."""
        return UserFlags.disable_premium.value

    @flag_value
    def quarantined(self):
        """:class:`bool`: Returns ``True`` if the user is quarantined."""
        return UserFlags.quarantined.value


@fill_with_flags()
class PremiumUsageFlags(BaseFlags):
    r"""Wraps up the Discord premium usage flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PremiumUsageFlags are equal.
        .. describe:: x != y

            Checks if two PremiumUsageFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PremiumUsageFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a PremiumUsageFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a PremiumUsageFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a PremiumUsageFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def premium_discriminator(self):
        """:class:`bool`: Returns ``True`` if the user has utilized premium discriminators."""
        return 1 << 0

    @flag_value
    def animated_avatar(self):
        """:class:`bool`: Returns ``True`` if the user has utilized animated avatars."""
        return 1 << 1

    @flag_value
    def profile_banner(self):
        """:class:`bool`: Returns ``True`` if the user has utilized profile banners."""
        return 1 << 2


@fill_with_flags()
class PurchasedFlags(BaseFlags):
    r"""Wraps up the Discord purchased flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PurchasedFlags are equal.
        .. describe:: x != y

            Checks if two PurchasedFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PurchasedFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a PurchasedFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a PurchasedFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a PurchasedFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def nitro_classic(self):
        """:class:`bool`: Returns ``True`` if the user has purchased Nitro classic."""
        return 1 << 0

    @flag_value
    def nitro(self):
        """:class:`bool`: Returns ``True`` if the user has purchased Nitro."""
        return 1 << 1

    @flag_value
    def guild_boost(self):
        """:class:`bool`: Returns ``True`` if the user has purchased a guild boost."""
        return 1 << 2

    @flag_value
    def nitro_basic(self):
        """:class:`bool`: Returns ``True`` if the user has purchased Nitro basic."""
        return 1 << 3


@fill_with_flags()
class MemberCacheFlags(BaseFlags):
    """Controls the library's cache policy when it comes to members.

    This allows for finer grained control over what members are cached.
    Note that the bot's own member is always cached. This class is passed
    to the ``member_cache_flags`` parameter in :class:`Client`.

    To construct an object you can pass keyword arguments denoting the flags
    to enable or disable.

    The default value is all flags enabled.

    .. versionadded:: 1.5

    .. container:: operations

        .. describe:: x == y

            Checks if two flags are equal.
        .. describe:: x != y

            Checks if two flags are not equal.

        .. describe:: x | y, x |= y

            Returns a MemberCacheFlags instance with all enabled flags from
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x & y, x &= y

            Returns a MemberCacheFlags instance with only flags enabled on
            both x and y.

            .. versionadded:: 2.0
        .. describe:: x ^ y, x ^= y

            Returns a MemberCacheFlags instance with only flags enabled on
            only one of x or y, not on both.

            .. versionadded:: 2.0
        .. describe:: ~x

            Returns a MemberCacheFlags instance with all flags inverted from x.

            .. versionadded:: 2.0
        .. describe:: hash(x)

               Return the flag's hash.
        .. describe:: iter(x)

               Returns an iterator of ``(name, value)`` pairs. This allows it
               to be, for example, constructed as a dict or a list of pairs.

    Attributes
    -----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    def __init__(self, **kwargs: bool):
        bits = max(self.VALID_FLAGS.values()).bit_length()
        self.value: int = (1 << bits) - 1
        for key, value in kwargs.items():
            if key not in self.VALID_FLAGS:
                raise TypeError(f"{key!r} is not a valid flag name.")
            setattr(self, key, value)

    @classmethod
    def all(cls: Type[MemberCacheFlags]) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with everything enabled."""
        bits = max(cls.VALID_FLAGS.values()).bit_length()
        value = (1 << bits) - 1
        self = cls.__new__(cls)
        self.value = value
        return self

    @classmethod
    def none(cls: Type[MemberCacheFlags]) -> MemberCacheFlags:
        """A factory method that creates a :class:`MemberCacheFlags` with everything disabled."""
        self = cls.__new__(cls)
        self.value = self.DEFAULT_VALUE
        return self

    @property
    def _empty(self):
        return self.value == self.DEFAULT_VALUE

    @flag_value
    def voice(self):
        """:class:`bool`: Whether to cache members that are in voice.

        Members that leave voice are no longer cached.
        """
        return 1

    @flag_value
    def other(self):
        """:class:`bool`: Whether to cache members that are collected from other means.

        This does not apply to members explicitly cached (e.g. :attr:`Guild.chunk`, :attr:`Guild.fetch_members`).

        There is an alias for this called :attr:`joined`.
        """
        return 2

    @alias_flag_value
    def joined(self):
        """:class:`bool`: Whether to cache members that are collected from other means.

        This does not apply to members explicitly cached (e.g. :attr:`Guild.chunk`, :attr:`Guild.fetch_members`).

        This is an alias for :attr:`other`.
        """
        return 2

    @property
    def _voice_only(self):
        return self.value == 1


@fill_with_flags()
class ApplicationFlags(BaseFlags):
    r"""Wraps up the Discord Application flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two ApplicationFlags are equal.
        .. describe:: x != y

            Checks if two ApplicationFlags are not equal.
        .. describe:: x | y, x |= y

            Returns an ApplicationFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns an ApplicationFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns an ApplicationFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns an ApplicationFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def embedded_released(self):
        """:class:`bool`: Returns ``True`` if the embedded application is released to the public."""
        return 1 << 1

    @flag_value
    def managed_emoji(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to create Twitch-style emotes."""
        return 1 << 2

    @flag_value
    def embedded_iap(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to use embedded in-app purchases."""
        return 1 << 3

    @flag_value
    def group_dm_create(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to create group DMs."""
        return 1 << 4

    @flag_value
    def rpc_private_beta(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to access the client RPC server."""
        return 1 << 5

    @flag_value
    def allow_assets(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to use activity assets."""
        return 1 << 8

    @flag_value
    def allow_activity_action_spectate(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to enable spectating activities."""
        return 1 << 9

    @flag_value
    def allow_activity_action_join_request(self):
        """:class:`bool`: Returns ``True`` if the application has the ability to enable activity join requests."""
        return 1 << 10

    @flag_value
    def rpc_has_connected(self):
        """:class:`bool`: Returns ``True`` if the application has accessed the client RPC server before."""
        return 1 << 11

    @flag_value
    def gateway_presence(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive presence information over the gateway.
        """
        return 1 << 12

    @flag_value
    def gateway_presence_limited(self):
        """:class:`bool`: Returns ``True`` if the application is allowed to receive
        presence information over the gateway.
        """
        return 1 << 13

    @flag_value
    def gateway_guild_members(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive full guild member lists.
        """
        return 1 << 14

    @flag_value
    def gateway_guild_members_limited(self):
        """:class:`bool`: Returns ``True`` if the application is allowed to receive full
        guild member lists.
        """
        return 1 << 15

    @flag_value
    def verification_pending_guild_limit(self):
        """:class:`bool`: Returns ``True`` if the application is currently pending verification
        and has hit the guild limit.
        """
        return 1 << 16

    @flag_value
    def embedded(self):
        """:class:`bool`: Returns ``True`` if the application is embedded within the Discord client."""
        return 1 << 17

    @flag_value
    def gateway_message_content(self):
        """:class:`bool`: Returns ``True`` if the application is verified and is allowed to
        receive message content."""
        return 1 << 18

    @flag_value
    def gateway_message_content_limited(self):
        """:class:`bool`: Returns ``True`` if the application is allowed to
        read message content in guilds."""
        return 1 << 19

    @flag_value
    def embedded_first_party(self):
        """:class:`bool`: Returns ``True`` if the embedded application is published by Discord."""
        return 1 << 20

    @flag_value
    def application_command_badge(self):
        """:class:`bool`: Returns ``True`` if the application has registered global application commands."""
        return 1 << 23

    @flag_value
    def active(self):
        """:class:`bool`: Returns ``True`` if the application is considered active.
        This means that it has had any global command executed in the past 30 days.
        """
        return 1 << 24


@fill_with_flags()
class ChannelFlags(BaseFlags):
    r"""Wraps up the Discord :class:`~discord.abc.GuildChannel` or :class:`Thread` flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two ChannelFlags are equal.
        .. describe:: x | y, x |= y

            Returns a ChannelFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a ChannelFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a ChannelFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a ChannelFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def pinned(self):
        """:class:`bool`: Returns ``True`` if the thread is pinned to the forum channel."""
        return 1 << 1

    @flag_value
    def require_tag(self):
        """:class:`bool`: Returns ``True`` if a tag is required to be specified when creating a thread in a :class:`ForumChannel`."""
        return 1 << 4


@fill_with_flags()
class PaymentSourceFlags(BaseFlags):
    r"""Wraps up the Discord payment source flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PaymentSourceFlags are equal.
        .. describe:: x != y

            Checks if two PaymentSourceFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PaymentSourceFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a PaymentSourceFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a PaymentSourceFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a PaymentSourceFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def new(self):
        """:class:`bool`: Returns ``True`` if the payment source is new."""
        return 1 << 0

    @flag_value
    def unknown(self):
        return 1 << 1


@fill_with_flags()
class SKUFlags(BaseFlags):
    r"""Wraps up the Discord SKU flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two SKUFlags are equal.
        .. describe:: x != y

            Checks if two SKUFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a SKUFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a SKUFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a SKUFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a SKUFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def premium_purchase(self):
        """:class:`bool`: Returns ``True`` if the SKU is a premium purchase."""
        return 1 << 0

    @flag_value
    def free_premium_content(self):
        """:class:`bool`: Returns ``True`` if the SKU is free premium content."""
        return 1 << 1

    @flag_value
    def available(self):
        """:class:`bool`: Returns ``True`` if the SKU is available for purchase."""
        return 1 << 2

    @flag_value
    def premium_and_distribution(self):
        """:class:`bool`: Returns ``True`` if the SKU is a premium or distribution product."""
        return 1 << 3

    @flag_value
    def sticker_pack(self):
        """:class:`bool`: Returns ``True`` if the SKU is a premium sticker pack."""
        return 1 << 4

    @flag_value
    def guild_role_subscription(self):
        """:class:`bool`: Returns ``True`` if the SKU is a guild role subscription. These are subscriptions made to guilds for premium perks."""
        return 1 << 5

    @flag_value
    def premium_subscription(self):
        """:class:`bool`: Returns ``True`` if the SKU is a Discord premium subscription or related first-party product.
        These are subscriptions like Nitro and Server Boosts. These are the only giftable subscriptions.
        """
        return 1 << 6

    @flag_value
    def application_guild_subscription(self):
        """:class:`bool`: Returns ``True`` if the SKU is a application subscription. These are subscriptions made to applications for premium perks bound to a guild."""
        return 1 << 7

    @flag_value
    def application_user_subscription(self):
        """:class:`bool`: Returns ``True`` if the SKU is a application subscription. These are subscriptions made to applications for premium perks bound to a user."""
        return 1 << 8


@fill_with_flags()
class PaymentFlags(BaseFlags):
    r"""Wraps up the Discord payment flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PaymentFlags are equal.
        .. describe:: x != y

            Checks if two PaymentFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PaymentFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a PaymentFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a PaymentFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a PaymentFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def gift(self):
        """:class:`bool`: Returns ``True`` if the payment is for a gift."""
        return 1 << 0

    @flag_value
    def preorder(self):
        """:class:`bool`: Returns ``True`` if the payment is a preorder."""
        return 1 << 3

    # TODO: The below are assumptions

    @flag_value
    def temporary_authorization(self):
        """:class:`bool`: Returns ``True`` if the payment is a temporary authorization."""
        return 1 << 5

    @flag_value
    def unknown(self):
        return 1 << 6


@fill_with_flags()
class PromotionFlags(BaseFlags):
    r"""Wraps up the Discord promotion flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two PromotionFlags are equal.
        .. describe:: x != y

            Checks if two PromotionFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a PromotionFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a PromotionFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a PromotionFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a PromotionFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def unknown_0(self):
        return 1 << 0

    @flag_value
    def unknown_1(self):
        # Possibly one month duration?
        return 1 << 1

    @flag_value
    def unknown_2(self):
        return 1 << 2

    @flag_value
    def unknown_3(self):
        return 1 << 3

    @flag_value
    def unknown_4(self):
        # Possibly unavailable/ended/inactive
        # Maybe also direct link
        # Maybe also available for existing users
        return 1 << 4

    @flag_value
    def blocked_ios(self):
        """:class:`bool`: Returns ``True`` if the promotion is blocked on iOS."""
        return 1 << 5

    @flag_value
    def outbound_redeemable_by_trial_users(self):
        """:class:`bool`: Returns ``True`` if the promotion is redeemable by trial users."""
        return 1 << 6


@fill_with_flags()
class GiftFlags(BaseFlags):
    r"""Wraps up the Discord payment flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two GiftFlags are equal.
        .. describe:: x != y

            Checks if two GiftFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a GiftFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a GiftFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a GiftFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a GiftFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def payment_source_required(self):
        """:class:`bool`: Returns ``True`` if the gift requires a payment source to redeem."""
        return 1 << 0

    @flag_value
    def existing_subscription_disallowed(self):
        """:class:`bool`: Returns ``True`` if the gift cannot be redeemed by users with existing premium subscriptions."""
        return 1 << 1

    @flag_value
    def not_self_redeemable(self):
        """:class:`bool`: Returns ``True`` if the gift cannot be redeemed by the gifter."""
        return 1 << 2

    # TODO: The below are assumptions

    @flag_value
    def promotion(self):
        """:class:`bool`: Returns ``True`` if the gift is from a promotion."""
        return 1 << 3


@fill_with_flags()
class LibraryApplicationFlags(BaseFlags):
    r"""Wraps up the Discord library application flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two LibraryApplicationFlags are equal.
        .. describe:: x != y

            Checks if two LibraryApplicationFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a LibraryApplicationFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a LibraryApplicationFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a LibraryApplicationFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a LibraryApplicationFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def hidden(self):
        """:class:`bool`: Returns ``True`` if the library application is hidden."""
        return 1 << 0

    @flag_value
    def private(self):
        """:class:`bool`: Returns ``True`` if the library application is not shown in playing status."""
        return 1 << 1

    @flag_value
    def overlay_disabled(self):
        """:class:`bool`: Returns ``True`` if the library application has the Discord overlay disabled."""
        return 1 << 2

    @flag_value
    def entitled(self):
        """:class:`bool`: Returns ``True`` if the library application is entitled to the user."""
        return 1 << 3

    @flag_value
    def premium(self):
        """:class:`bool`: Returns ``True`` if the library application is free for premium users."""
        return 1 << 4


@fill_with_flags()
class ApplicationDiscoveryFlags(BaseFlags):
    r"""Wraps up the Discord application discovery eligibility flags.

    .. container:: operations

        .. describe:: x == y

            Checks if two LibraryApplicationFlags are equal.
        .. describe:: x != y

            Checks if two LibraryApplicationFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a LibraryApplicationFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a LibraryApplicationFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a LibraryApplicationFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a LibraryApplicationFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def verified(self):
        """:class:`bool`: Returns ``True`` if the application is verified."""
        return 1 << 0

    @flag_value
    def tag(self):
        """:class:`bool`: Returns ``True`` if the application has at least one tag set."""
        return 1 << 1

    @flag_value
    def description(self):
        """:class:`bool`: Returns ``True`` if the application has a description."""
        return 1 << 2

    @flag_value
    def terms_of_service(self):
        """:class:`bool`: Returns ``True`` if the application has a terms of service."""
        return 1 << 3

    @flag_value
    def privacy_policy(self):
        """:class:`bool`: Returns ``True`` if the application has a privacy policy."""
        return 1 << 4

    @flag_value
    def install_params(self):
        """:class:`bool`: Returns ``True`` if the application has a custom install URL or install parameters."""
        return 1 << 5

    @flag_value
    def safe_name(self):
        """:class:`bool`: Returns ``True`` if the application name is safe for work."""
        return 1 << 6

    @flag_value
    def safe_description(self):
        """:class:`bool`: Returns ``True`` if the application description is safe for work."""
        return 1 << 7

    @flag_value
    def approved_commands(self):
        """:class:`bool`: Returns ``True`` if the application has the message content intent approved or utilizes application commands."""
        return 1 << 8

    @flag_value
    def support_guild(self):
        """:class:`bool`: Returns ``True`` if the application has a support guild set."""
        return 1 << 9

    @flag_value
    def safe_commands(self):
        """:class:`bool`: Returns ``True`` if the application's commands are safe for work."""
        return 1 << 10

    @flag_value
    def mfa(self):
        """:class:`bool`: Returns ``True`` if the application's owner has MFA enabled."""
        return 1 << 11

    @flag_value
    def safe_directory_overview(self):
        """:class:`bool`: Returns ``True`` if the application's directory long description is safe for work."""
        return 1 << 12

    @flag_value
    def supported_locales(self):
        """:class:`bool`: Returns ``True`` if the application has at least one supported locale set."""
        return 1 << 13

    @flag_value
    def safe_short_description(self):
        """:class:`bool`: Returns ``True`` if the application's directory short description is safe for work."""
        return 1 << 14

    @flag_value
    def safe_role_connections(self):
        """:class:`bool`: Returns ``True`` if the application's role connections metadata is safe for work."""
        return 1 << 15

    @flag_value
    def eligible(self):
        """:class:`bool`: Returns ``True`` if the application has met all the above criteria and is eligible for discovery."""
        return 1 << 16


@fill_with_flags()
class FriendSourceFlags(BaseFlags):
    r"""Wraps up the Discord friend source flags.

    These are used in user settings to control who can add you as a friend.

    .. container:: operations

        .. describe:: x == y

            Checks if two FriendSourceFlags are equal.
        .. describe:: x != y

            Checks if two FriendSourceFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a FriendSourceFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a FriendSourceFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a FriendSourceFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a FriendSourceFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @classmethod
    def _from_dict(cls, data: dict) -> Self:
        self = cls()
        if data.get("mutual_friends"):
            self.mutual_friends = True
        if data.get("mutual_guilds"):
            self.mutual_guilds = True
        if data.get("all"):
            self.no_relation = True
        return self

    def _to_dict(self) -> dict:
        return {
            "mutual_friends": self.mutual_friends,
            "mutual_guilds": self.mutual_guilds,
            "all": self.no_relation,
        }

    @classmethod
    def none(cls) -> Self:
        """A factory method that creates a :class:`FriendSourceFlags` that allows no friend request."""
        return cls()

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates a :class:`FriendSourceFlags` that allows any friend requests."""
        self = cls()
        self.no_relation = True
        return self

    @flag_value
    def mutual_friends(self):
        """:class:`bool`: Returns ``True`` if a user can add you as a friend if you have mutual friends."""
        return 1 << 1

    @flag_value
    def mutual_guilds(self):
        """:class:`bool`: Returns ``True`` if a user can add you as a friend if you are in the same guild."""
        return 1 << 2

    @flag_value
    def no_relation(self):
        """:class:`bool`: Returns ``True`` if a user can always add you as a friend."""
        # Requires all of the above
        return 1 << 3 | 1 << 2 | 1 << 1


@fill_with_flags()
class FriendDiscoveryFlags(BaseFlags):
    r"""Wraps up the Discord friend discovery flags.

    These are used in user settings to control how you get recommended friends.

    .. container:: operations

        .. describe:: x == y

            Checks if two FriendDiscoveryFlags are equal.
        .. describe:: x != y

            Checks if two FriendDiscoveryFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a FriendDiscoveryFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a FriendDiscoveryFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a FriendDiscoveryFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a FriendDiscoveryFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @classmethod
    def none(cls) -> Self:
        """A factory method that creates a :class:`FriendDiscoveryFlags` that allows no friend discovery."""
        return cls()

    @classmethod
    def all(cls) -> Self:
        """A factory method that creates a :class:`FriendDiscoveryFlags` that allows all friend discovery."""
        self = cls()
        self.find_by_email = True
        self.find_by_phone = True
        return self

    @flag_value
    def find_by_phone(self):
        """:class:`bool`: Returns ``True`` if a user can add you as a friend if they have your phone number."""
        return 1 << 1

    @flag_value
    def find_by_email(self):
        """:class:`bool`: Returns ``True`` if a user can add you as a friend if they have your email address."""
        return 1 << 2


@fill_with_flags()
class HubProgressFlags(BaseFlags):
    """Wraps up the Discord hub progress flags.

    These are used in user settings, specifically guild progress, to track engagement and feature usage in hubs.

    .. container:: operations

        .. describe:: x == y

            Checks if two HubProgressFlags are equal.
        .. describe:: x != y

            Checks if two HubProgressFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a HubProgressFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a HubProgressFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a HubProgressFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a HubProgressFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def join_guild(self):
        """:class:`bool`: Returns ``True`` if the user has joined a guild in the hub."""
        return 1 << 0

    @flag_value
    def invite_user(self):
        """:class:`bool`: Returns ``True`` if the user has sent an invite for the hub."""
        return 1 << 1

    @flag_value
    def contact_sync(self):
        """:class:`bool`: Returns ``True`` if the user has accepted the contact sync modal."""
        return 1 << 2


@fill_with_flags()
class OnboardingProgressFlags(BaseFlags):
    """Wraps up the Discord guild onboarding progress flags.

    These are used in user settings, specifically guild progress, to track engagement and feature usage in guild onboarding.

    .. container:: operations

        .. describe:: x == y

            Checks if two OnboardingProgressFlags are equal.
        .. describe:: x != y

            Checks if two OnboardingProgressFlags are not equal.
        .. describe:: x | y, x |= y

            Returns a OnboardingProgressFlags instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns a OnboardingProgressFlags instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns a OnboardingProgressFlags instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns a OnboardingProgressFlags instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. This value is a bit array field of a 53-bit integer
        representing the currently available flags. You should query
        flags via the properties rather than using this raw value.
    """

    __slots__ = ()

    @flag_value
    def notice_shown(self):
        """:class:`bool`: Returns ``True`` if the user has been shown the onboarding notice."""
        return 1 << 0

    @flag_value
    def notice_cleared(self):
        """:class:`bool`: Returns ``True`` if the user has cleared the onboarding notice."""
        return 1 << 1


@fill_with_flags()
class AutoModPresets(ArrayFlags):
    r"""Wraps up the Discord :class:`AutoModRule` presets.

    .. container:: operations

        .. describe:: x == y

            Checks if two AutoModPresets flags are equal.
        .. describe:: x != y

            Checks if two AutoModPresets flags are not equal.
        .. describe:: x | y, x |= y

            Returns an AutoModPresets instance with all enabled flags from
            both x and y.
        .. describe:: x & y, x &= y

            Returns an AutoModPresets instance with only flags enabled on
            both x and y.
        .. describe:: x ^ y, x ^= y

            Returns an AutoModPresets instance with only flags enabled on
            only one of x or y, not on both.
        .. describe:: ~x

            Returns an AutoModPresets instance with all flags inverted from x.
        .. describe:: hash(x)

            Return the flag's hash.
        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.

    .. versionadded:: 2.0

    Attributes
    -----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    @classmethod
    def all(cls: Type[Self]) -> Self:
        """A factory method that creates a :class:`AutoModPresets` with everything enabled."""
        bits = max(cls.VALID_FLAGS.values()).bit_length()
        value = (1 << bits) - 1
        self = cls.__new__(cls)
        self.value = value
        return self

    @classmethod
    def none(cls: Type[Self]) -> Self:
        """A factory method that creates a :class:`AutoModPresets` with everything disabled."""
        self = cls.__new__(cls)
        self.value = self.DEFAULT_VALUE
        return self

    @flag_value
    def profanity(self):
        """:class:`bool`: Whether to use the preset profanity filter."""
        return 1 << 0

    @flag_value
    def sexual_content(self):
        """:class:`bool`: Whether to use the preset sexual content filter."""
        return 1 << 1

    @flag_value
    def slurs(self):
        """:class:`bool`: Whether to use the preset slurs filter."""
        return 1 << 2


@fill_with_flags()
class MemberFlags(BaseFlags):
    r"""Wraps up the Discord Guild Member flags

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two MemberFlags are equal.

        .. describe:: x != y

            Checks if two MemberFlags are not equal.

        .. describe:: x | y, x |= y

            Returns a MemberFlags instance with all enabled flags from
            both x and y.

        .. describe:: x & y, x &= y

            Returns a MemberFlags instance with only flags enabled on
            both x and y.

        .. describe:: x ^ y, x ^= y

            Returns a MemberFlags instance with only flags enabled on
            only one of x or y, not on both.

        .. describe:: ~x

            Returns a MemberFlags instance with all flags inverted from x.

        .. describe:: hash(x)

            Return the flag's hash.

        .. describe:: iter(x)

            Returns an iterator of ``(name, value)`` pairs. This allows it
            to be, for example, constructed as a dict or a list of pairs.
            Note that aliases are not shown.


    Attributes
    -----------
    value: :class:`int`
        The raw value. You should query flags via the properties
        rather than using this raw value.
    """

    @flag_value
    def did_rejoin(self):
        """:class:`bool`: Returns ``True`` if the member left and rejoined the :attr:`~discord.Member.guild`."""
        return 1 << 0

    @flag_value
    def completed_onboarding(self):
        """:class:`bool`: Returns ``True`` if the member has completed onboarding."""
        return 1 << 1

    @flag_value
    def bypasses_verification(self):
        """:class:`bool`: Returns ``True`` if the member can bypass the guild verification requirements."""
        return 1 << 2

    @flag_value
    def started_onboarding(self):
        """:class:`bool`: Returns ``True`` if the member has started onboarding."""
        return 1 << 3



from discord_self._vendor.discord.types.channel import PartialChannel
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.user import User
from discord_self._vendor.discord.types.webhook import (
    FollowerWebhook,
    PartialWebhook,
    SourceGuild,
    Webhook,
    WebhookType,
)

__all__ = [
    "FollowerWebhook",
    "PartialChannel",
    "PartialWebhook",
    "Snowflake",
    "SourceGuild",
    "User",
    "Webhook",
    "WebhookType",
]



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
    runtime_checkable,
)

from .enums import (
    AppCommandOptionType,
    AppCommandType,
    ChannelType,
    InteractionType,
    try_enum,
)
from .interactions import _wrapped_interaction
from .mixins import Hashable
from .permissions import Permissions
from .utils import _generate_nonce, _get_as_snowflake

if TYPE_CHECKING:
    from .abc import Messageable, Snowflake
    from .application import IntegrationApplication
    from .file import File
    from .guild import Guild
    from .interactions import Interaction
    from .message import Attachment, Message
    from .state import ConnectionState


__all__ = (
    "BaseCommand",
    "UserCommand",
    "MessageCommand",
    "SlashCommand",
    "SubCommand",
    "Option",
    "OptionChoice",
)


@runtime_checkable
class ApplicationCommand(Protocol):
    """An ABC that represents a usable application command.

    The following implement this ABC:

    - :class:`~discord.UserCommand`
    - :class:`~discord.MessageCommand`
    - :class:`~discord.SlashCommand`
    - :class:`~discord.SubCommand`

    .. versionadded:: 2.0

    Attributes
    -----------
    name: :class:`str`
        The command's name.
    description: :class:`str`
        The command's description, if any.
    type: :class:`~discord.AppCommandType`
        The type of application command.
    default_permission: :class:`bool`
        Whether the command is enabled in guilds by default.
    dm_permission: :class:`bool`
        Whether the command is enabled in DMs.
    nsfw: :class:`bool`
        Whether the command is marked NSFW and only available in NSFW channels.
    application: Optional[:class:`~discord.IntegrationApplication`]
        The application this command belongs to.
        Only available if requested.
    application_id: :class:`int`
        The ID of the application this command belongs to.
    guild_id: Optional[:class:`int`]
        The ID of the guild this command is registered in. A value of ``None``
        denotes that it is a global command.
    """

    __slots__ = ()

    if TYPE_CHECKING:
        _state: ConnectionState
        _channel: Optional[Messageable]
        _default_member_permissions: Optional[int]
        name: str
        description: str
        version: int
        type: AppCommandType
        default_permission: bool
        dm_permission: bool
        nsfw: bool
        application_id: int
        application: Optional[IntegrationApplication]
        mention: str
        guild_id: Optional[int]

    def __str__(self) -> str:
        return self.name

    async def __call__(
        self,
        data: dict,
        files: Optional[List[File]] = None,
        channel: Optional[Messageable] = None,
    ) -> Interaction:
        channel = channel or self.target_channel
        if channel is None:
            raise TypeError("__call__() missing 1 required argument: 'channel'")

        return await _wrapped_interaction(
            self._state,
            _generate_nonce(),
            InteractionType.application_command,
            data["name"],
            await channel._get_channel(),  # type: ignore # acc_channel is always correct here
            data,
            files=files,
            application_id=self.application_id,
        )

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`~discord.Guild`]: Returns the guild this command is registered to
        if it exists.
        """
        return self._state._get_guild(self.guild_id)

    def is_group(self) -> bool:
        """Query whether this command is a group.

        Returns
        -------
        :class:`bool`
            Whether this command is a group.
        """
        return False

    @property
    def target_channel(self) -> Optional[Messageable]:
        """Optional[:class:`.abc.Messageable`]: The channel this application command will be used on.

        You can set this in order to use this command in a different channel without re-fetching it.
        """
        return self._channel

    @target_channel.setter
    def target_channel(self, value: Optional[Messageable]) -> None:
        from .abc import Messageable

        if not isinstance(value, Messageable) and value is not None:
            raise TypeError("channel must derive from Messageable")
        self._channel = value

    @property
    def default_member_permissions(self) -> Optional[Permissions]:
        """Optional[:class:`~discord.Permissions`]: The default permissions required to use this command.

        .. note::
            This may be overrided on a guild-by-guild basis.
        """
        perms = self._default_member_permissions
        return Permissions(perms) if perms is not None else None


class BaseCommand(ApplicationCommand, Hashable):
    __slots__ = (
        "name",
        "description",
        "id",
        "version",
        "type",
        "default_permission",
        "application",
        "application_id",
        "dm_permission",
        "nsfw",
        "guild_id",
        "_data",
        "_state",
        "_channel",
        "_default_member_permissions",
    )

    def __init__(
        self,
        *,
        state: ConnectionState,
        data: Dict[str, Any],
        channel: Optional[Messageable] = None,
        **kwargs,
    ) -> None:
        self._state = state
        self._data = data
        self.name = data["name"]
        self.description = data["description"]
        self._channel = channel
        self.application_id: int = int(data["application_id"])
        self.id: int = int(data["id"])
        self.version = int(data["version"])
        self.type = try_enum(AppCommandType, data["type"])

        application = data.get("application")
        self.application = (
            state.create_integration_application(application) if application else None
        )

        self._default_member_permissions = _get_as_snowflake(
            data, "default_member_permissions"
        )
        self.default_permission: bool = data.get("default_permission", True)
        dm_permission = data.get("dm_permission")  # Null means true?
        self.dm_permission = dm_permission if dm_permission is not None else True
        self.nsfw: bool = data.get("nsfw", False)
        self.guild_id: Optional[int] = _get_as_snowflake(data, "guild_id")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name!r}>"

    @property
    def mention(self) -> str:
        """:class:`str`: Returns a string that allows you to mention the command."""
        return f"</{self.name}:{self.id}>"


class SlashMixin(ApplicationCommand, Protocol):
    if TYPE_CHECKING:
        _parent: SlashCommand
        options: List[Option]
        children: List[SubCommand]

    async def __call__(
        self,
        options: List[dict],
        files: Optional[List[File]],
        attachments: List[Attachment],
        channel: Optional[Messageable] = None,
    ) -> Interaction:
        obj = self._parent
        command = obj._data
        command["name_localized"] = command["name"]
        data = {
            "application_command": command,
            "attachments": attachments,
            "id": str(obj.id),
            "name": obj.name,
            "options": options,
            "type": obj.type.value,
            "version": str(obj.version),
        }
        if self.guild_id:
            data["guild_id"] = str(self.guild_id)
        return await super().__call__(data, files, channel)

    def _parse_kwargs(
        self, kwargs: Dict[str, Any]
    ) -> Tuple[List[Dict[str, Any]], List[File], List[Attachment]]:
        possible_options = {o.name: o for o in self.options}
        kwargs = {k: v for k, v in kwargs.items() if k in possible_options}
        options = []
        files = []

        for k, v in kwargs.items():
            option = possible_options[k]
            type = option.type

            if type in {
                AppCommandOptionType.user,
                AppCommandOptionType.channel,
                AppCommandOptionType.role,
                AppCommandOptionType.mentionable,
            }:
                v = str(v.id)
            elif type is AppCommandOptionType.boolean:
                v = bool(v)
            elif type is AppCommandOptionType.attachment:
                files.append(v)
                v = len(files) - 1
            else:
                v = option._convert(v)

            if type is AppCommandOptionType.string:
                v = str(v)
            elif type is AppCommandOptionType.integer:
                v = int(v)
            elif type is AppCommandOptionType.number:
                v = float(v)

            options.append({"name": k, "value": v, "type": type.value})

        attachments = []
        for index, file in enumerate(files):
            attachments.append(file.to_dict(index))

        return options, files, attachments

    def _unwrap_options(self, data: List[Dict[str, Any]]) -> None:
        options = []
        children = []
        for option in data:
            type = try_enum(AppCommandOptionType, option["type"])
            if type in {
                AppCommandOptionType.sub_command,
                AppCommandOptionType.sub_command_group,
            }:
                children.append(SubCommand(parent=self, data=option))
            else:
                options.append(Option(option))

        self.options = options
        self.children = children


class UserCommand(BaseCommand):
    """Represents a user command.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two commands are equal.

        .. describe:: x != y

            Checks if two commands are not equal.

        .. describe:: hash(x)

            Return the command's hash.

        .. describe:: str(x)

            Returns the command's name.

    Attributes
    ----------
    id: :class:`int`
        The command's ID.
    version: :class:`int`
        The command's version.
    name: :class:`str`
        The command's name.
    description: :class:`str`
        The command's description, if any.
    type: :class:`AppCommandType`
        The type of application command. This will always be :attr:`AppCommandType.user`.
    default_permission: :class:`bool`
        Whether the command is enabled in guilds by default.
    dm_permission: :class:`bool`
        Whether the command is enabled in DMs.
    nsfw: :class:`bool`
        Whether the command is marked NSFW and only available in NSFW channels.
    application: Optional[:class:`IntegrationApplication`]
        The application this command belongs to.
        Only available if requested.
    application_id: :class:`int`
        The ID of the application this command belongs to.
    guild_id: Optional[:class:`int`]
        The ID of the guild this command is registered in. A value of ``None``
        denotes that it is a global command.
    """

    __slots__ = ("_user",)

    def __init__(self, *, target: Optional[Snowflake] = None, **kwargs):
        super().__init__(**kwargs)
        self._user = target

    async def __call__(
        self, user: Optional[Snowflake] = None, *, channel: Optional[Messageable] = None
    ):
        """|coro|

        Use the user command.

        Parameters
        ----------
        user: Optional[:class:`User`]
            The user to use the command on. Overrides :attr:`target_user`.
            Required if :attr:`target_user` is not set.
        channel: Optional[:class:`abc.Messageable`]
            The channel to use the command on. Overrides :attr:`target_channel`.
            Required if :attr:`target_channel` is not set.

        Returns
        -------
        :class:`Interaction`
            The interaction that was created.
        """
        user = user or self._user
        if user is None:
            raise TypeError("__call__() missing 1 required positional argument: 'user'")

        command = self._data
        data = {
            "application_command": command,
            "attachments": [],
            "id": str(self.id),
            "name": self.name,
            "options": [],
            "target_id": str(user.id),
            "type": self.type.value,
            "version": str(self.version),
        }
        return await super().__call__(data, None, channel)

    @property
    def target_user(self) -> Optional[Snowflake]:
        """Optional[:class:`~abc.Snowflake`]: The user this application command will be used on.

        You can set this in order to use this command on a different user without re-fetching it.
        """
        return self._user

    @target_user.setter
    def target_user(self, value: Optional[Snowflake]) -> None:
        from .abc import Snowflake

        if not isinstance(value, Snowflake) and value is not None:
            raise TypeError("user must be Snowflake")
        self._user = value


class MessageCommand(BaseCommand):
    """Represents a message command.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two commands are equal.

        .. describe:: x != y

            Checks if two commands are not equal.

        .. describe:: hash(x)

            Return the command's hash.

        .. describe:: str(x)

            Returns the command's name.

    Attributes
    ----------
    id: :class:`int`
        The command's ID.
    version: :class:`int`
        The command's version.
    name: :class:`str`
        The command's name.
    description: :class:`str`
        The command's description, if any.
    type: :class:`AppCommandType`
        The type of application command. This will always be :attr:`AppCommandType.message`.
    default_permission: :class:`bool`
        Whether the command is enabled in guilds by default.
    dm_permission: :class:`bool`
        Whether the command is enabled in DMs.
    nsfw: :class:`bool`
        Whether the command is marked NSFW and only available in NSFW channels.
    application: Optional[:class:`IntegrationApplication`]
        The application this command belongs to.
        Only available if requested.
    application_id: :class:`int`
        The ID of the application this command belongs to.
    guild_id: Optional[:class:`int`]
        The ID of the guild this command is registered in. A value of ``None``
        denotes that it is a global command.
    """

    __slots__ = ("_message",)

    def __init__(self, *, target: Optional[Message] = None, **kwargs):
        super().__init__(**kwargs)
        self._message = target

    async def __call__(
        self,
        message: Optional[Message] = None,
        *,
        channel: Optional[Messageable] = None,
    ):
        """|coro|

        Use the message command.

        Parameters
        ----------
        message: Optional[:class:`Message`]
            The message to use the command on. Overrides :attr:`target_message`.
            Required if :attr:`target_message` is not set.
        channel: Optional[:class:`abc.Messageable`]
            The channel to use the command on. Overrides :attr:`target_channel`.
            Required if :attr:`target_channel` is not set.

        Returns
        -------
        :class:`Interaction`
            The interaction that was created.
        """
        message = message or self._message
        if message is None:
            raise TypeError(
                "__call__() missing 1 required positional argument: 'message'"
            )

        command = self._data
        command["name_localized"] = command["name"]
        data = {
            "application_command": command,
            "attachments": [],
            "id": str(self.id),
            "name": self.name,
            "options": [],
            "target_id": str(message.id),
            "type": self.type.value,
            "version": str(self.version),
        }
        return await super().__call__(data, None, channel)

    @property
    def target_message(self) -> Optional[Message]:
        """Optional[:class:`Message`]: The message this application command will be used on.

        You can set this in order to use this command on a different message without re-fetching it.
        """
        return self._message

    @target_message.setter
    def target_message(self, value: Optional[Message]) -> None:
        from .message import Message

        if not isinstance(value, Message) and value is not None:
            raise TypeError("message must be Message")
        self._message = value


class SlashCommand(BaseCommand, SlashMixin):
    """Represents a slash command.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two commands are equal.

        .. describe:: x != y

            Checks if two commands are not equal.

        .. describe:: hash(x)

            Return the command's hash.

        .. describe:: str(x)

            Returns the command's name.

    Attributes
    ----------
    id: :class:`int`
        The command's ID.
    version: :class:`int`
        The command's version.
    name: :class:`str`
        The command's name.
    description: :class:`str`
        The command's description, if any.
    type: :class:`AppCommandType`
        The type of application command.
    default_permission: :class:`bool`
        Whether the command is enabled in guilds by default.
    dm_permission: :class:`bool`
        Whether the command is enabled in DMs.
    nsfw: :class:`bool`
        Whether the command is marked NSFW and only available in NSFW channels.
    application: Optional[:class:`IntegrationApplication`]
        The application this command belongs to.
        Only available if requested.
    application_id: :class:`int`
        The ID of the application this command belongs to.
    guild_id: Optional[:class:`int`]
        The ID of the guild this command is registered in. A value of ``None``
        denotes that it is a global command.
    options: List[:class:`Option`]
        The command's options.
    children: List[:class:`SubCommand`]
        The command's subcommands. If a command has subcommands, it is a group and cannot be used.
    """

    __slots__ = ("_parent", "options", "children")

    def __init__(self, *, data: Dict[str, Any], **kwargs) -> None:
        super().__init__(data=data, **kwargs)
        self._parent = self
        self._unwrap_options(data.get("options", []))

    async def __call__(self, channel: Optional[Messageable] = None, /, **kwargs):
        r"""|coro|

        Use the slash command.

        Parameters
        ----------
        channel: Optional[:class:`abc.Messageable`]
            The channel to use the command on. Overrides :attr:`target_channel`.
            Required if :attr:`target_channel` is not set.
        \*\*kwargs: Any
            The options to use. These will be casted to the correct type.
            If an option has choices, they are automatically converted from name to value for you.

        Raises
        ------
        TypeError
            Attempted to use a group.

        Returns
        -------
        :class:`Interaction`
            The interaction that was created.
        """
        if self.is_group():
            raise TypeError("Cannot use a group")

        return await super().__call__(*self._parse_kwargs(kwargs), channel)

    def __repr__(self) -> str:
        BASE = f"<SlashCommand id={self.id} name={self.name!r}"
        if self.options:
            BASE += f" options={len(self.options)}"
        if self.children:
            BASE += f" children={len(self.children)}"
        return BASE + ">"

    def is_group(self) -> bool:
        """Query whether this command is a group.

        Returns
        -------
        :class:`bool`
            Whether this command is a group.
        """
        return bool(self.children)


class SubCommand(SlashMixin):
    """Represents a slash command child.

    .. versionadded:: 2.0

    This could be a subcommand, or a subgroup.

    .. container:: operations

        .. describe:: str(x)

            Returns the command's name.

    Attributes
    ----------
    name: :class:`str`
        The subcommand's name.
    description: :class:`str`
        The subcommand's description, if any.
    type: :class:`AppCommandType`
        The type of application command. Always :attr:`AppCommandType.chat_input`.
    parent: Union[:class:`SlashCommand`, :class:`SubCommand`]
        The parent command.
    options: List[:class:`Option`]
        The subcommand's options.
    children: List[:class:`SubCommand`]
        The subcommand's subcommands. If a subcommand has subcommands, it is a group and cannot be used.
    """

    __slots__ = (
        "_parent",
        "_state",
        "_type",
        "parent",
        "options",
        "children",
        "type",
    )

    def __init__(self, *, parent, data):
        self.name = data["name"]
        self.description = data.get("description")
        self._state = parent._state
        self.parent: Union[SlashCommand, SubCommand] = parent
        self._parent: SlashCommand = getattr(parent, "parent", parent)  # type: ignore
        self.type = AppCommandType.chat_input  # Avoid confusion I guess
        self._type: AppCommandOptionType = try_enum(AppCommandOptionType, data["type"])
        self._unwrap_options(data.get("options", []))

    def __str__(self) -> str:
        return self.name

    def _walk_parents(self):
        parent = self.parent
        while True:
            if isinstance(parent, SlashCommand):
                break
            else:
                yield parent
                parent = parent.parent

    async def __call__(self, channel: Optional[Messageable] = None, /, **kwargs):
        r"""|coro|

        Use the sub command.

        Parameters
        ----------
        channel: Optional[:class:`abc.Messageable`]
            The channel to use the command on. Overrides :attr:`target_channel`.
            Required if :attr:`target_channel` is not set.
        \*\*kwargs: Any
            The options to use. These will be casted to the correct type.
            If an option has choices, they are automatically converted from name to value for you.

        Raises
        ------
        TypeError
            Attempted to use a group.

        Returns
        -------
        :class:`Interaction`
            The interaction that was created.
        """
        if self.is_group():
            raise TypeError("Cannot use a group")

        options, files, attachments = self._parse_kwargs(kwargs)

        options = [
            {
                "type": self._type.value,
                "name": self.name,
                "options": options,
            }
        ]
        for parent in self._walk_parents():
            options = [
                {
                    "type": parent._type.value,
                    "name": parent.name,
                    "options": options,
                }
            ]

        return await super().__call__(options, files, attachments, channel)

    def __repr__(self) -> str:
        BASE = f"<SubCommand name={self.name!r}"
        if self.options:
            BASE += f" options={len(self.options)}"
        if self.children:
            BASE += f" children={len(self.children)}"
        return BASE + ">"

    @property
    def qualified_name(self) -> str:
        """:class:`str`: Returns the fully qualified command name.
        The qualified name includes the parent name as well. For example,
        in a command like ``/foo bar`` the qualified name is ``foo bar``.
        """
        names = [self.name, self.parent.name]
        if isinstance(self.parent, SubCommand):
            names.append(self._parent.name)
        return " ".join(reversed(names))

    @property
    def mention(self) -> str:
        """:class:`str`: Returns a string that allows you to mention the subcommand."""
        return f"</{self.qualified_name}:{self._parent.id}>"

    @property
    def _default_member_permissions(self) -> Optional[int]:
        return self._parent._default_member_permissions

    @property
    def application_id(self) -> int:
        """:class:`int`: The ID of the application this command belongs to."""
        return self._parent.application_id

    @property
    def version(self) -> int:
        """:class:`int`: The version of the command."""
        return self._parent.version

    @property
    def default_permission(self) -> bool:
        """:class:`bool`: Whether the command is enabled in guilds by default."""
        return self._parent.default_permission

    @property
    def dm_permission(self) -> bool:
        """:class:`bool`: Whether the command is enabled in DMs."""
        return self._parent.dm_permission

    @property
    def nsfw(self) -> bool:
        """:class:`bool`: Whether the command is marked NSFW and only available in NSFW channels."""
        return self._parent.nsfw

    @property
    def guild_id(self) -> Optional[int]:
        """Optional[:class:`int`]: The ID of the guild this command is registered in. A value of ``None``
        denotes that it is a global command."""
        return self._parent.guild_id

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`~discord.Guild`]: Returns the guild this command is registered to
        if it exists.
        """
        return self._parent.guild

    def is_group(self) -> bool:
        """Query whether this command is a group.

        Returns
        -------
        :class:`bool`
            Whether this command is a group.
        """
        return self._type is AppCommandOptionType.sub_command_group

    @property
    def application(self):
        """Optional[:class:`IntegrationApplication`]: The application this command belongs to.
        Only available if requested.
        """
        return self._parent.application

    @property
    def target_channel(self) -> Optional[Messageable]:
        """Optional[:class:`.abc.Messageable`]: The channel this command will be used on.

        You can set this in order to use this command on a different channel without re-fetching it.
        """
        return self._parent.target_channel

    @target_channel.setter
    def target_channel(self, value: Optional[Messageable]) -> None:
        self._parent.target_channel = value


class Option:
    """Represents a command option.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: str(x)

            Returns the option's name.

    Attributes
    ----------
    name: :class:`str`
        The option's name.
    description: :class:`str`
        The option's description, if any.
    type: :class:`AppCommandOptionType`
        The type of option.
    required: :class:`bool`
        Whether the option is required.
    min_value: Optional[Union[:class:`int`, :class:`float`]]
        Minimum value of the option. Only applicable to :attr:`AppCommandOptionType.integer` and :attr:`AppCommandOptionType.number`.
    max_value: Optional[Union[:class:`int`, :class:`float`]]
        Maximum value of the option. Only applicable to :attr:`AppCommandOptionType.integer` and :attr:`AppCommandOptionType.number`.
    choices: List[:class:`OptionChoice`]
        A list of possible choices to choose from. If these are present, you must choose one from them.

        Only applicable to :attr:`AppCommandOptionType.string`, :attr:`AppCommandOptionType.integer`, and :attr:`AppCommandOptionType.number`.
    channel_types: List[:class:`ChannelType`]
        A list of channel types that you can choose from. If these are present, you must choose a channel that is one of these types.

        Only applicable to :attr:`AppCommandOptionType.channel`.
    autocomplete: :class:`bool`
        Whether the option autocompletes.

        Only applicable to :attr:`AppCommandOptionType.string`, :attr:`AppCommandOptionType.integer`, and :attr:`AppCommandOptionType.number`.
        Always ``False`` if :attr:`choices` are present.
    """

    __slots__ = (
        "name",
        "description",
        "type",
        "required",
        "min_value",
        "max_value",
        "choices",
        "channel_types",
        "autocomplete",
    )

    def __init__(self, data):
        self.name: str = data["name"]
        self.description: str = data["description"]
        self.type: AppCommandOptionType = try_enum(AppCommandOptionType, data["type"])
        self.required: bool = data.get("required", False)
        self.min_value: Optional[Union[int, float]] = data.get("min_value")
        self.max_value: Optional[int] = data.get("max_value")
        self.choices = [
            OptionChoice(choice, self.type) for choice in data.get("choices", [])
        ]
        self.channel_types: List[ChannelType] = [
            try_enum(ChannelType, c) for c in data.get("channel_types", [])
        ]
        self.autocomplete: bool = data.get("autocomplete", False)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return (
            f"<Option name={self.name!r} type={self.type!r} required={self.required}>"
        )

    def _convert(self, value):
        for choice in self.choices:
            if (new_value := choice._convert(value)) != value:
                return new_value
        return value


class OptionChoice:
    """Represents a choice for an option.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: str(x)

            Returns the choice's name.

    Attributes
    ----------
    name: :class:`str`
        The choice's displayed name.
    value: Union[:class:`str`, :class:`int`, :class:`float`]
        The choice's value. The type of this depends on the option's type.
    """

    __slots__ = ("name", "value")

    def __init__(self, data: Dict[str, str], type: AppCommandOptionType):
        self.name: str = data["name"]
        self.value: Union[str, int, float]
        if type is AppCommandOptionType.string:
            self.value = data["value"]
        elif type is AppCommandOptionType.integer:
            self.value = int(data["value"])
        elif type is AppCommandOptionType.number:
            self.value = float(data["value"])

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<OptionChoice name={self.name!r} value={self.value!r}>"

    def _convert(self, value):
        if value == self.name:
            return self.value
        return value


def _command_factory(command_type: int) -> Tuple[AppCommandType, Type[BaseCommand]]:
    value = try_enum(AppCommandType, command_type)
    if value is AppCommandType.chat_input:
        return value, SlashCommand
    elif value is AppCommandType.user:
        return value, UserCommand
    elif value is AppCommandType.message:
        return value, MessageCommand
    else:
        return value, BaseCommand  # IDK about this
